"""
GUI for CDAWeb.

A GUI that can download data files from CDAWeb
and load them into pytplot variables.

Requires cdasws.
For cdasws documentation, see:
    https://pypi.org/project/cdasws/
    https://cdaweb.gsfc.nasa.gov/WebServices/REST/py/cdasws/index.html


To open the GUI window:
    from pyspedas.cdagui.cdagui import cdagui
    x = cdagui()
To start the gui from the command line, use:
    python pyspedas/cdagui/cdagui.py
"""
import os
import time
import tkinter as tk
from tkinter import messagebox, filedialog
from cdaweb import CDAWeb
from config import CONFIG


class cdaWindow:

    def __init__(self, master):
        # CDAWeb connection
        cda = CDAWeb()

        # Defaults
        default_start_time = "2023-01-01 00:00:00"
        default_end_time = "2023-01-01 23:59:59"
        download_box = tk.IntVar()
        default_dir = CONFIG["local_data_dir"]
        default_status = "Status: Ready!"

        def status(txt):
            # Change status bar, show wait cursor
            if txt == "":
                status_label.config(text=default_status)
                master.config(cursor="")
            else:
                status_label.config(text="Status: " + str(txt))
                master.config(cursor="wait")
                time.sleep(0.2)

            master.update_idletasks()
            master.update()

        def select_dir():
            # Button: Select dir
            initial_dir = str(dir_entry.get())
            path = filedialog.askdirectory(parent=window, initialdir=initial_dir, title="Please select a directory", mustexist=True)
            if os.path.exists(path):
                dir_entry.delete(0, tk.END)
                dir_entry.insert(0, path)

        def clear_boxes():
            # Button: Clear all list boxes
            mission_list.selection_clear(0, tk.END)
            instrument_list.selection_clear(0, tk.END)
            dataset_list.delete(0, tk.END)
            file_list.delete(0, tk.END)
            label_groups.config(text="Selected Mission Groups:")
            label_instruments.config(text="Selected Instrument Types:")

        def exit_gui():
            # Button: Exit
            window.destroy()

        def find_datasets():
            # Button: 1. Find Datasets
            msgtitle = "Find Datasets"

            # Change status
            status(msgtitle + "...")

            datasets = []
            dataset_list.delete(0, tk.END)
            file_list.delete(0, tk.END)
            sel_g = list(mission_list.curselection())
            sel_g_val = [mission_list.get(index) for index in sel_g]
            sel_i = list(instrument_list.curselection())
            sel_i_val = [instrument_list.get(index) for index in sel_i]
            if len(sel_g_val) < 1:
                messagebox.showerror(msgtitle, "Please select one or more Mission Groups!")
            elif len(sel_i_val) < 1:
                messagebox.showerror(msgtitle, "Please select one or more Instrument Types!")
            else:
                label_groups.config(text="Selected Mission Groups: [" + ",".join(sel_g_val) + "]")
                label_instruments.config(text="Selected Instrument Types: [" + ",".join(sel_i_val) + "]")

                datasets = cda.get_datasets(sel_g_val, sel_i_val)
                if len(datasets) < 1:
                    messagebox.showerror(msgtitle, "No datasets found for these parameters!")
                else:
                    for i in datasets:
                        dataset_list.insert(tk.END, str(i))

            status("")  # Reset status

            return datasets

        def find_filelist():
            # Button: 2. Get File List
            msgtitle = "Get File List"

            # Change status
            status(msgtitle + "...")

            file_list.delete(0, tk.END)
            all_files = []
            sel_d = list(dataset_list.curselection())
            sel_d_val = [dataset_list.get(index) for index in sel_d]
            if len(sel_d_val) < 1:
                messagebox.showerror(msgtitle, "Please select one or more Datasets!")
                status("")  # Reset status
                return []

            t0 = str(start_time.get())
            t1 = str(end_time.get())
            if len(t0) < 1 or len(t1) < 1:
                messagebox.showerror(msgtitle, "Please set start and end time!")
                status("")  # Reset status
                return []

            # Get all files for this dataset and start/end times.
            all_files = cda.get_filenames(sel_d_val, t0, t1)
            filelen = len(all_files)
            if filelen < 1:
                msg = "No files were found with these parameters."
                messagebox.showinfo(msgtitle, msg)
                status("")  # Reset status
                return []
            elif filelen > 50:
                msg = "Number of files found: " + str(filelen)
                msg += "\nOnly 50 will be shown."
                messagebox.showinfo(msgtitle, msg)

            # Add filenames to listbox (up to 50 files)
            for i in all_files[:50]:
                file_list.insert(tk.END, str(i))

            status("")  # Reset status

            # Return all files, even if they are more than 50
            return all_files

        def get_data():
            # Button: 3. Get Data
            msgtitle = "Get Data"

            # Change status
            status(msgtitle + "...")

            file_result = []
            sel_f = list(file_list.curselection())
            sel_f_val = [file_list.get(index) for index in sel_f]
            if len(sel_f_val) < 1:
                messagebox.showerror(msgtitle, "Please select one or more files to download!")
                status("")  # Reset status
                return []

            if download_box.get() == 1:
                download_only = True
            else:
                download_only = False

            local_dir = str(dir_entry.get())
            if len(local_dir) < 1:
                local_dir = default_dir
                dir_entry.insert(0, default_dir)

            # Get the files, and/or the tplot variables
            # sesults is a list [remote filename, local filename, status]
            result = cda.cda_download(sel_f_val, local_dir, download_only)

            # Show a message about the results
            filelen = len(result)
            if filelen < 1:
                msg = "No files were downloaded."
                messagebox.showerror(msgtitle, msg)
                status("")  # Reset status
                return []
            else:
                file_result = result
                count_no_downloads = 0
                count_tplot_problem = 0
                count_tplot = 0
                for item in result:
                    if len(item) != 3:
                        continue
                    if item[2] == -1:
                        count_no_downloads += 1
                    elif item[2] == 0 and not download_only:
                        count_tplot_problem += 1
                    elif item[2] == 1:
                        count_tplot += 1
                msg = "Results:"
                msg += "\n"
                msg += "\nFiles to download: " + str(filelen)
                msg += ("\nFiles that could not be downloaded: "
                        + str(count_no_downloads))
                if not download_only:
                    msg += "\n"
                    msg += ("\nFiles loaded to pytplot: " + str(count_tplot))
                    msg += ("\nFiles that could not be loaded to pytplot: "
                            + str(count_tplot_problem))
                # Show final message with results
                messagebox.showinfo(msgtitle, msg)

            status("")  # Reset status
            return file_result

        # Create the main window
        window = master
        window.title("CDAWeb Data Downloader")
        window.geometry("800x600")
        window.configure()
        window.option_add("*font", "lucida 10")
        window.state("zoomed")  # Start maximazed
        window.minsize(800, 600)

        # Size of grid
        no_of_cols = 2
        no_of_rows = 14

        # Create columns
        for i in range(no_of_cols):
            window.grid_columnconfigure(i, weight=1)

        # Create rows
        for i in range(no_of_rows):
            window.grid_rowconfigure(i, weight=1)

        # Row 0 - just a label
        window.grid_rowconfigure(0, weight=0)
        label00 = tk.Label(window, text="Download Data from CDAWeb", bg="#AFEEEE", font=("Helvetica", 10, "bold"))
        label00.grid(row=0, columnspan=2, sticky="new")

        # Row 1, 2 - Mission Groups and Instrument Types
        window.grid_rowconfigure(1, weight=0)
        label10 = tk.Label(window, text="Mission Groups:")
        label10.grid(row=1, column=0, sticky="ws")

        label11 = tk.Label(window, text="Instrument Types:")
        label11.grid(row=1, column=1, sticky="ws")

        cell20 = tk.Frame(window)
        cell20.grid(row=2, column=0, sticky="nsew")
        mission_list = tk.Listbox(cell20, selectmode=tk.MULTIPLE, exportselection=False)
        mission_list.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=5, ipadx=5, ipady=5, expand=True)
        cdagroups = cda.get_observatories()
        # cdagroups = [i for i in range(200)]
        for g in cdagroups:
            mission_list.insert(tk.END, str(g))
        scrollbar20 = tk.Scrollbar(mission_list, orient="vertical")
        scrollbar20.pack(side=tk.RIGHT, fill=tk.BOTH)
        mission_list.config(yscrollcommand=scrollbar20.set)
        scrollbar20.config(command=mission_list.yview)

        cell21 = tk.Frame(window)
        cell21.grid(row=2, column=1, sticky="nsew")
        instrument_list = tk.Listbox(cell21, selectmode=tk.MULTIPLE, exportselection=False)
        instrument_list.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=5, ipadx=5, ipady=5, expand=True)
        cdainstr = cda.get_instruments()
        # cdainstr = [i for i in range(200)]
        for i in cdainstr:
            instrument_list.insert(tk.END, str(i))
        scrollbar21 = tk.Scrollbar(instrument_list, orient="vertical")
        scrollbar21.pack(side=tk.RIGHT, fill=tk.BOTH)
        instrument_list.config(yscrollcommand=scrollbar21.set)
        scrollbar21.config(command=instrument_list.yview)

        # Row 3 - Find Datasets button
        window.grid_rowconfigure(3, weight=0)
        msg = "Select Mission Group(s) and Instrument Type(s) and press:"
        label30 = tk.Label(window, text=msg)
        label30.grid(row=3, column=0, sticky="nse")

        button31 = tk.Button(window, text="1. Find Datasets", bg="#AFEEAF", command=find_datasets, width=30)
        button31.grid(row=3, column=1, sticky="nsw")

        # Row 4 - Dataset labels
        window.grid_rowconfigure(4, weight=0)
        cell40 = tk.Frame(window)
        cell40.grid(row=4, columnspan=2, sticky="new")
        for i in range(3):
            cell40.grid_rowconfigure(i, weight=1)
        cell40.grid_columnconfigure(0, weight=1)
        label_groups = tk.Label(cell40, text="Selected Mission Groups:", justify="left", anchor="w")
        label_groups.grid(row=0, sticky="new")
        label_instruments = tk.Label(cell40, text="Selected Instrument Types:", justify="left", anchor="w")
        label_instruments.grid(row=1, sticky="new")
        label_2 = tk.Label(cell40, text="Datasets:", justify="left", anchor="w")
        label_2.grid(row=2, sticky="new")

        # Row 5 - Datasets
        cell50 = tk.Frame(window)
        cell50.grid(row=5, columnspan=2, sticky="nsew")
        dataset_list = tk.Listbox(cell50, selectmode=tk.MULTIPLE, exportselection=False)
        dataset_list.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=5, ipadx=5, ipady=5, expand=True)
        scrollbar50 = tk.Scrollbar(dataset_list, orient="vertical")
        scrollbar50.pack(side=tk.RIGHT, fill=tk.BOTH)
        dataset_list.config(yscrollcommand=scrollbar50.set)
        scrollbar50.config(command=dataset_list.yview)

        # Row 6, 7 - Date Time
        window.grid_rowconfigure(6, weight=0)
        label60 = tk.Label(window, text="Date and Time (format: YYYY-MM-DD[ HH:MM:SS])", justify="left", anchor="w")
        label60.grid(row=6, columnspan=2, sticky="new")

        window.grid_rowconfigure(7, weight=0)
        cell70 = tk.Frame(window)
        cell70.grid(row=7, column=0, sticky="new")
        cell70.grid_rowconfigure(0, weight=1)
        cell70.grid_columnconfigure(0, weight=0)
        cell70.grid_columnconfigure(1, weight=1)
        cell70.grid_columnconfigure(2, weight=0)
        label70 = tk.Label(cell70, text="Start Time:", justify="left", anchor="w")
        label70.grid(row=0, column=0, sticky="wsn")
        start_time = tk.Entry(cell70)
        start_time.insert(0, default_start_time)
        start_time.grid(row=0, column=1, sticky="ewsn", padx=6)

        cell71 = tk.Frame(window)
        cell71.grid(row=7, column=1, sticky="new")
        cell71.grid_rowconfigure(0, weight=1)
        cell71.grid_columnconfigure(0, weight=0)
        cell71.grid_columnconfigure(1, weight=1)
        cell71.grid_columnconfigure(2, weight=0)
        label71 = tk.Label(cell71, text="End Time:", justify="left", anchor="w")
        label71.grid(row=0, column=0, sticky="wsn")
        end_time = tk.Entry(cell71)
        end_time.insert(0, default_end_time)
        end_time.grid(row=0, column=1, sticky="sewn", padx=6)

        # Row 8 - Get File List button
        window.grid_rowconfigure(8, weight=0)
        msg = "Select Dataset(s) and Times and press:"
        label80 = tk.Label(window, text=msg)
        label80.grid(row=8, column=0, sticky="nse", pady=6)

        button81 = tk.Button(window, text="2. Get File List", bg="#AFEEAF", command=find_filelist, width=30)
        button81.grid(row=8, column=1, sticky="nsw", pady=6)

        # Row 9, 10 - Files
        window.grid_rowconfigure(9, weight=0)
        label90 = tk.Label(window, text="Remote Files:", justify="left", anchor="w")
        label90.grid(row=9, columnspan=2, sticky="new")

        cell100 = tk.Frame(window)
        cell100.grid(row=10, columnspan=2, sticky="nsew")
        file_list = tk.Listbox(cell100, selectmode=tk.MULTIPLE, exportselection=False)
        file_list.pack(side=tk.TOP, fill=tk.BOTH, padx=5, pady=5, ipadx=5, ipady=5, expand=True)
        scrollbar100 = tk.Scrollbar(file_list, orient="vertical")
        scrollbar100.pack(side=tk.RIGHT, fill=tk.BOTH)
        file_list.config(yscrollcommand=scrollbar100.set)
        scrollbar100.config(command=file_list.yview)

        # Row 11 - Download Directory
        window.grid_rowconfigure(11, weight=0)
        cell110 = tk.Frame(window)
        cell110.grid(row=11, columnspan=2, sticky="new")
        cell110.grid_columnconfigure(0, weight=0)
        cell110.grid_columnconfigure(1, weight=1)
        cell110.grid_columnconfigure(2, weight=0)

        label110 = tk.Label(cell110, text="Local Directory:", justify="left", anchor="w")
        label110.grid(row=0, column=0, sticky="new", padx=4)
        dir_entry = tk.Entry(cell110)
        dir_entry.insert(0, default_dir)
        dir_entry.grid(row=0, column=1, sticky="sewn", padx=4)
        # dir_entry.bind("<Key>", lambda e: "break")  # Make it read-only
        button110 = tk.Button(cell110, text="Select Directory", command=select_dir, width=15)
        button110.grid(row=0, column=2, sticky="wns", padx=4)

        # Row 12 - Download Only checkbox, Get Data button, Clear, Exit
        window.grid_rowconfigure(12, weight=0)

        cell120 = tk.Frame(window)
        cell120.grid(row=12, column=0, sticky="new", pady=4)
        cell120.grid_columnconfigure(0, weight=0)
        cell120.grid_columnconfigure(1, weight=1)
        cell120.grid_columnconfigure(2, weight=0)
        only_ch = tk.Checkbutton(cell120, text="Download Only", variable=download_box, onvalue=1, offvalue=0)
        only_ch.grid(row=0, column=0, sticky="new", padx=4)
        msg = "Select Files(s) and press:"
        label120 = tk.Label(cell120, text=msg)
        label120.grid(row=0, column=2, sticky="nse", pady=6)

        cell121 = tk.Frame(window)
        cell121.grid(row=12, column=1, sticky="new", pady=4)
        cell121.grid_columnconfigure(0, weight=0)
        cell121.grid_columnconfigure(1, weight=1)
        cell121.grid_columnconfigure(2, weight=0)
        cell121.grid_columnconfigure(3, weight=0)
        button122 = tk.Button(cell121, text="3. Get Data", bg="#AFEEAF", command=get_data, width=30)
        button122.grid(row=0, column=0, sticky="wns", padx=4)
        button123 = tk.Button(cell121, text="Clear", command=clear_boxes, width=10, bg="#E9967A")
        button123.grid(row=0, column=2, sticky="wns", padx=4)
        button124 = tk.Button(cell121, text="Exit", command=exit_gui, width=10, bg="#E9967A")
        button124.grid(row=0, column=3, sticky="wns", padx=4)

        # Row 13 - Status bar
        window.grid_rowconfigure(13, weight=0)
        status_label = tk.Label(window, text=default_status, justify="left", anchor="w", relief="groove")
        status_label.grid(row=13, columnspan=2, sticky="new", padx=2)


def cdagui():
    root = tk.Tk()
    startgui = cdaWindow(root)
    root.mainloop()
    return startgui


if __name__ == "__main__":
    root = tk.Tk()
    startgui = cdaWindow(root)
    root.mainloop()



import math


class SpinmodelSegment:
    """ An object describing a single segment of a spin model.

    Attributes:
        t1 (float): Start time of segment (seconds since epoch)
        t2 (float): End time of segment (seconds since epoch)
        c1 (int): Spin count at start of segment
        c2 (int): Spin count at end of segment
        b (float): Initial spin rate in degrees/sec
        c (float): Acceleration of spin rate in degrees/sec^2
        npts (int): Number of data points used to construct this segment
        maxgap (float): Largest gap, in seconds, in data used to construct this segment
        phaserr (float): The maximum fitting error in this segment (seconds)
        initial_delta_phi (float): The offset in degrees between the corrected spin model and uncorrected IDPU model at
            the start of the segment
        idpu_spinper (float): The onboard spin period during an eclipse segment
        segflags (int): Bit mapped flags describing the eclipse and correction status during this segment
            Bit 0 (lsb) eclipse flag, bit 1 waveform corrections applied, bit 2 spin fit corrections applied
    """

    __slots__ = 't1', 't2', 'c1', 'c2', 'b', 'c', 'npts', 'maxgap', 'phaserr', 'initial_delta_phi', 'idpu_spinper', \
                'segflags'

    def __init__(self,
                 t1: float,
                 t2: float,
                 c1: int,
                 c2: int,
                 b: float,
                 c: float,
                 npts: int,
                 maxgap: float,
                 phaserr: float,
                 initial_delta_phi: float,
                 idpu_spinper: float,
                 segflags: int):
        self.t1 = t1
        self.t2 = t2
        self.c1 = c1
        self.c2 = c2
        self.b = b
        self.c = c
        self.npts = npts
        self.maxgap = maxgap
        self.phaserr = phaserr
        self.initial_delta_phi = initial_delta_phi
        self.idpu_spinper = idpu_spinper
        self.segflags = segflags

    def print(self):
        """ Print a segment using sensible formatting

        Args:

        Returns:
         """

        print('%20.8f %20.8f %d %d %f %e %d %f %f %f %f %d' % (self.t1, self.t2, self.c1, self.c2,
                                                               self.b, self.c, self.npts, self.maxgap, self.phaserr,
                                                               self.initial_delta_phi,
                                                               self.idpu_spinper, self.segflags))

# Previously, this module included a set of interpolation and extrapolation routines for
# single spinmodel segments:
#
# extrap_before_t()
# extrap_after_t()
# extrap_before_n()
# extrap_after_n()
#
# These routines have been inlined where needed, and users should be calling the interp_t or interp_n methods on
# the spinmodel object.   If needed, the removed routines can be restored from the version history, or the IDL code.
# JWL 2023-03-13
#

    def interp_t(self,
                 t: float) -> (float, float, int, float, float):
        """Return modeled values for a time falling within the segment start/end times

        Args:
            t (float): Input time (seconds since epoch)

        Returns:
            A tuple containing the modeled spin period, spin count, spin phase, eclipse_delta_phi, and t_last
                at the input time.

        """

        dt = t - self.t1
        phi = (self.b * dt + self.c * dt * dt)
        bp = self.b + 2.0 * self.c * dt
        spinper = 360.0 / bp
        spincount, spinphase = divmod(phi, 360.0)
        fracspins = phi / 360.0
        phi_lastpulse = spincount * 360.0
        if abs(self.c) < 1.0e-12:
            tlast_dt = phi_lastpulse / self.b
        else:
            b = self.b
            c = self.c
            tlast_dt = (-b + math.sqrt(b * b - 4.0 * c * (-phi_lastpulse))) / (2.0 * c)
        if ((self.segflags & 3) == 3) and (self.idpu_spinper > 1.0):
            model_phi = fracspins * 360.0
            idpu_bp = 360.0 / self.idpu_spinper
            idpu_phi = dt * idpu_bp
            eclipse_delta_phi = self.initial_delta_phi + (model_phi - idpu_phi)

        else:
            eclipse_delta_phi = 0.0
        t_last = self.t1 + tlast_dt
        spincount = spincount + self.c1
        return spinper, spinphase, spincount, eclipse_delta_phi, t_last






import cdflib
from pytplot import clip, options, ylim, zlim, get_data

from ..load import load


def pwe_hfa(trange=['2017-04-01', '2017-04-02'],
            datatype='spec',
            mode='low',
            level='l2',
            suffix='',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            uname=None,
            passwd=None,
            time_clip=False,
            ror=True):
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        ror: bool
            If set, print PI info and rules of the road

    Returns:
        List of tplot variables created.

    """

    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24

    if level == 'l2':
        prefix = 'erg_pwe_hfa_'+level+'_' + mode + '_'

    if level == 'l2':
        pathformat = 'satellite/erg/pwe/hfa/'+level+'/'+datatype+'/'+mode + \
            '/%Y/%m/erg_pwe_hfa_'+level+'_'+datatype+'_'+mode+'_%Y%m%d_v??_??.cdf'
    elif level == 'l3':
        prefix = 'erg_pwe_hfa_'+level+'_1min_'
        pathformat = 'satellite/erg/pwe/hfa/'+level + \
            '/%Y/%m/erg_pwe_hfa_'+level+'_1min_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)

    if (len(loaded_data) > 0) and ror:

        try:
            if isinstance(loaded_data, list):
                if downloadonly:
                    cdf_file = cdflib.CDF(loaded_data[-1])
                    gatt = cdf_file.globalattsget()
                else:
                    gatt = get_data(loaded_data[-1], metadata=True)['CDF']['GATT']
            elif isinstance(loaded_data, dict):
                gatt = loaded_data[list(loaded_data.keys())[-1]]['CDF']['GATT']

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE HFA')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: "+gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/HFA: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Hfa')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if (level == 'l2') and (mode == 'low') and (not notplot):

        # set spectrogram plot option
        options(prefix + 'spectra_eu' + suffix, 'Spec', 1)
        options(prefix + 'spectra_ev' + suffix, 'Spec', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'Spec', 1)
        options(prefix + 'spectra_esum' + suffix, 'Spec', 1)
        options(prefix + 'spectra_er' + suffix, 'Spec', 1)
        options(prefix + 'spectra_el' + suffix, 'Spec', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'Spec', 1)
        options(prefix + 'spectra_e_ar' + suffix, 'Spec', 1)

        if prefix + 'spectra_er' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_er' + suffix, 0., 5000.)
        if prefix + 'spectra_el' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_el' + suffix, 0., 5000.)

        if prefix + 'spectra_eu' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_eu' + suffix, 0., 5000.)
            # set ylim
            ylim(prefix + 'spectra_eu' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_eu' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_ev' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_ev' + suffix, 0., 5000.)
            # set ylim
            ylim(prefix + 'spectra_ev' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_ev' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_bgamma' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_bgamma' + suffix, 2.0, 200.0)
            # set zlim
            zlim(prefix + 'spectra_bgamma' + suffix, 1e-4, 1e+2)

        if prefix + 'spectra_esum' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_esum' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_esum' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_e_ar' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_e_ar' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_e_ar' + suffix, -1, 1)

        # set y axis to logscale
        options(prefix + 'spectra_eu' + suffix, 'ylog', 1)
        options(prefix + 'spectra_ev' + suffix, 'ylog', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'ylog', 1)
        options(prefix + 'spectra_esum' + suffix, 'ylog', 1)
        options(prefix + 'spectra_er' + suffix, 'ylog', 1)
        options(prefix + 'spectra_el' + suffix, 'ylog', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'ylog', 1)
        options(prefix + 'spectra_e_ar' + suffix, 'ylog', 1)

        # set z axis to logscale
        options(prefix + 'spectra_eu' + suffix, 'zlog', 1)
        options(prefix + 'spectra_ev' + suffix, 'zlog', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'zlog', 1)
        options(prefix + 'spectra_esum' + suffix, 'zlog', 1)
        options(prefix + 'spectra_er' + suffix, 'zlog', 1)
        options(prefix + 'spectra_el' + suffix, 'zlog', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'zlog', 1)

        # set ytitle
        options(prefix + 'spectra_eu' + suffix, 'ytitle', 'ERG PWE/HFA (EU)')
        options(prefix + 'spectra_ev' + suffix, 'ytitle', 'ERG PWE/HFA (EV)')
        options(prefix + 'spectra_esum' + suffix,
                'ytitle', 'ERG PWE/HFA (ESUM)')
        options(prefix + 'spectra_e_ar' + suffix,
                'ytitle', 'ERG PWE/HFA (E_AR)')
        options(prefix + 'spectra_bgamma' + suffix,
                'ytitle', 'ERG PWE/HFA (BGAMMA)')

        # set ysubtitle
        options(prefix + 'spectra_eu' + suffix, 'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_ev' + suffix, 'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_esum' + suffix,
                'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_e_ar' + suffix,
                'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_bgamma' + suffix,
                'ysubtitle', 'frequency [Hz]')

        # set ztitle
        options(prefix + 'spectra_eu' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_ev' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_esum' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_e_ar' + suffix, 'ztitle', 'LH:-1/RH:+1')
        options(prefix + 'spectra_bgamma' + suffix, 'ztitle', 'pT^2/Hz')

        # change colormap option
        options(prefix + 'spectra_eu' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_ev' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_bgamma' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_esum' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_er' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_el' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_e_mix' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_e_ar' + suffix, 'Colormap', 'jet')

    elif level == 'l3':

        # set ytitle
        options(prefix + 'Fuhr' + suffix, 'ytitle', 'UHR frequency [Mhz]')
        options(prefix + 'ne_mgf' + suffix,
                'ytitle', 'eletctorn density [/cc]')

        # set y axis to logscale
        options(prefix + 'Fuhr' + suffix, 'ylog', 1)
        options(prefix + 'ne_mgf' + suffix, 'ylog', 1)

    return loaded_data



"""Test fuctions in the utilites folder."""
import unittest

from pyspedas.utilities.dailynames import dailynames
from pyspedas import tcopy
from pytplot import data_exists, tkm2re
from pytplot import get_data, store_data, options


class UtilTestCases(unittest.TestCase):
    """Test fuctions in the utilites folder."""

    def test_dailynames(self):
        """Test dailynames function."""
        self.assertTrue(dailynames(trange=['2015-12-1', '2015-12-1/2:00'],
                                   hour_res=True) == ['2015120100',
                                                      '2015120101'])
        self.assertTrue(dailynames(trange=['2015-12-1', '2015-12-3'])
                        == ['20151201', '20151202'])
        self.assertTrue(dailynames(trange=['2015-12-3', '2015-12-2'])
                        == ['20151203'])
        self.assertTrue(dailynames() is None)
        self.assertTrue(dailynames(trange=['2015-12-3', '2019-12-2'],
                                   file_format='%Y') ==
                        ['2015', '2016', '2017', '2018', '2019'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-3-2'],
                                   file_format='%Y%m') == ['201501', '201502',
                                                           '201503'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-3-2'],
                                   file_format='/%Y/%m/') ==
                        ['/2015/01/', '/2015/02/', '/2015/03/'])
        self.assertTrue(dailynames(trange=['2015-1-1', '2015-1-1/3:00'],
                                   file_format='%H', res=60.0) ==
                        ['00', '01', '02'])
        self.assertTrue(dailynames(trange=['2015-1-1/2:00', '2015-1-1/3:00'],
                                   file_format='%M', res=600.) ==
                        ['00', '10', '20', '30', '40', '50'])

    def test_tcopy(self):
        """Test tcopy function."""
        store_data('test', data={'x': [1, 2, 3], 'y': [5, 5, 5]})
        tcopy('test')
        tcopy('test', 'another-copy')
        t, d = get_data('test-copy')
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        t, d = get_data('another-copy')
        self.assertTrue(t.tolist() == [1, 2, 3])
        self.assertTrue(d.tolist() == [5, 5, 5])
        # the following should gracefully error
        tcopy('doesnt exist', 'another-copy')
        tcopy(['another-copy', 'test'], 'another-copy')

    def test_tkm2re(self):
        store_data('test', data={'x': [1, 2, 3], 'y': [5, 5, 5]})
        options('test', 'ysubtitle', '[Re]')
        # convert to km
        tkm2re('test', km=True)
        # convert back
        tkm2re('test_km')
        self.assertTrue(data_exists('test_km_re'))
        nothing = tkm2re('doesnt_exist')
        self.assertTrue(nothing is None)
        tkm2re('test_km', newname='another_test_km')
        self.assertTrue(data_exists('another_test_km'))
        anerror = tkm2re('test_km', newname=['test1_km', 'test1_km'])
        self.assertTrue(anerror is None)


if __name__ == '__main__':
    unittest.main()



"""
Subtracts the median from the data.

Notes
-----
Similar to tsub_average.pro in IDL SPEDAS.

"""
import logging
import pytplot

def subtract_median(names, new_names=None, suffix=None, overwrite=None):
    """
    Subtracts the median from data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-d'.
    overwrite: bool, optional
        If set, then pytplot variables are replaced.

    Returns
    -------
    None.

    """
    logging.info("subtract_median has been moved to the pytplot.tplot_math package. Please update your imports!")
    logging.info("This version will eventually be removed.")

    pytplot.tplot_math.subtract_median(names=names, new_names=new_names, suffix=suffix, overwrite=overwrite)



import logging
import unittest
import pyspedas
import pytplot
from numpy.testing import assert_allclose

class TplotTimeValidation(unittest.TestCase):
    """ Tests creation of support variables in themis.state() """

    @classmethod
    def setUpClass(cls):
        pass

    def setUp(self):
        pass
        """ We need to clean tplot variables before each run"""


    def test_timespan(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,timespan
        from pyspedas.erg import mgf
        vars = mgf(trange=['2017-03-27', '2017-03-28'])  # load MGF Lv.2 8-s data for 0-24 UT on Mar. 27, 2017.
        tplot('erg_mgf_l2_mag_8sec_sm')
        timespan('2017-03-27 09:00:00', 6, keyword='hours')
        tplot(['erg_mgf_l2_mag_8sec_sm', 'erg_mgf_l2_mag_8sec_gsm'])

    def test_subsec_timespan(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,timespan
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl') # full plot
        timespan('2007-03-23/14:00',0.9,"seconds")
        tplot('tha_fgl_dsl')  # short time interval

    def test_subsec_tlimit(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        tlimit(['2007-03-23/14:00','2007-03-23/14:00:00.9'])
        tplot('tha_fgl_dsl')  # short time interval

    def test_timebar(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit, timebar
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        # Test various timebar formats
        # Standard format
        timebar('2007-03-23 14:00:00')
        tplot('tha_fgl_dsl')
        # Slash between date and time
        timebar('2007-03-23/13:00:00')
        tplot('tha_fgl_dsl')
        # ISO8601
        timebar('20070323T1330')
        tplot('tha_fgl_dsl')
        tlimit(['2007-03-23/14:00', '2007-03-23/14:00:00.9'])
        # Subsecond precision
        timebar('2007-03-23/14:00:00.5')
        tplot('tha_fgl_dsl')

    def test_tlimit_full_arg(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        tlimit(['2007-03-23/14:00','2007-03-23/14:00:00.9'])
        tplot('tha_fgl_dsl')  # short time interval
        tlimit('full')
        tplot('tha_fgl_dsl')  # back to full interval

    def test_tlimit_full_flag(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        tlimit(['2007-03-23/14:00','2007-03-23/14:00:00.9'])
        tplot('tha_fgl_dsl')  # short time interval
        tlimit(full=True)
        tplot('tha_fgl_dsl')  # back to full interval

    def test_tlimit_last_arg(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        tlimit(['2007-03-23/14:00','2007-03-23/15:00:00'])
        tplot('tha_fgl_dsl')  # time interval 1  14:00 to 15:00
        tlimit(['2007-03-23/12:00','2007-03-23/13:00:00'])
        tplot('tha_fgl_dsl')  # time interval 2 12:00 to 13:00
        tlimit('last')
        tplot('tha_fgl_dsl')  # back to time interval 1, 14:00 to 15:00

    def test_tlimit_last_flag(self):
        """Test pytplot.timespan as used in ERG notebook"""
        from pytplot import tplot,tlimit
        from pyspedas.themis import fgm
        vars = fgm(probe='a',level='l2',trange=['2007-03-23', '2007-03-24'])
        tplot('tha_fgl_dsl')  # full plot
        tlimit(['2007-03-23/14:00','2007-03-23/15:00:00'])
        tplot('tha_fgl_dsl')  # time interval 1  14:00 to 15:00
        tlimit(['2007-03-23/12:00','2007-03-23/13:00:00'])
        tplot('tha_fgl_dsl')  # time interval 2 12:00 to 13:00
        tlimit(last=True)
        tplot('tha_fgl_dsl')  # back to time interval 1, 14:00 to 15:00


if __name__ == '__main__':
    unittest.main()



import os
import csv
import time
import logging
import pickle
import numpy as np
from pytplot import time_double, time_string
from pyspedas.mms.mms_login_lasp import mms_login_lasp
from pyspedas.utilities.download import download
from pyspedas.mms.mms_config import CONFIG
from pyspedas.mms.mms_tai2unix import mms_tai2unix


def mms_update_brst_intervals():
    """
    This function downloads and caches the current mms_burst_data_segment.csv
    file from the MMS SDC
    """
    # not sure if logging in is still important for these
    # so this code might be unnecessary now; for now it
    # remains to match the IDL functionality
    login = mms_login_lasp()

    if login is None:
        logging.error('Error logging into the LASP SDC.')
        return

    session, user = login

    # grab ~6 months of burst intervals at a time
    start_interval = '2015-03-01'
    end_interval = time_double(start_interval) + 6.*30*24*60*60

    unix_starts = []
    unix_ends = []

    while time_double(start_interval) <= time_double(time.time()):
        start_str = time_string(time_double(start_interval), fmt='%d-%b-%Y')
        end_str = time_string(end_interval, fmt='%d-%b-%Y')

        logging.info(f'Downloading updates for {start_str} - {end_str}')

        remote_path = 'https://lasp.colorado.edu/mms/sdc/public/service/latis/'
        remote_file = f'mms_burst_data_segment.csv?FINISHTIME>={start_str}+&FINISHTIME<{end_str}'

        brst_file = download(remote_path=remote_path, remote_file=remote_file,
                             local_file=os.path.join(CONFIG['local_data_dir'], 'mms_burst_data_segment.csv'),
                             session=session, no_wildcards=True)

        if isinstance(brst_file, list):
            # should only be one file
            brst_file = brst_file[0]

        times = load_csv_file(brst_file)
        if not isinstance(times, tuple) or len(times) != 3:
            logging.error('Error loading the CSV file')
            return

        taistarttime, taiendtime, status = times

        complete_idxs = np.argwhere(status == 'COMPLETE+FINISHED').flatten()
        if len(complete_idxs) != 0:
            tai_starts = taistarttime[complete_idxs]
            tai_ends = taiendtime[complete_idxs]

            unix_starts.extend(mms_tai2unix(tai_starts))
            unix_ends.extend(mms_tai2unix(tai_ends))

        logging.info(f'Done grabbing updates for {start_str}-{end_str}')

        start_interval = end_interval
        end_interval = time_double(start_interval) + 6. * 30 * 24 * 60 * 60

    brst_intervals = {'start_times': unix_starts,
                      'end_times': unix_ends}

    with open(os.path.join(CONFIG['local_data_dir'], 'mms_brst_intervals.pickle'), "wb") as file:
        pickle.dump(brst_intervals, file)

    logging.info(f'Burst intervals updated! Last interval in the file: {time_string(unix_starts[-1])}-{time_string(unix_ends[-1])}')

    return brst_intervals


def load_csv_file(filename):
    taistarttime = []
    taiendtime = []
    status = []
    with open(filename, 'r') as file:
        reader = csv.reader(file)
        next(reader)  # skip the header row
        for row in reader:
            taistarttime.append(int(row[1]))
            taiendtime.append(int(row[2]))
            status.append(row[7])
    return np.array(taistarttime), np.array(taiendtime), np.array(status)



from .version import version
from pytplot import data_exists
from pytplot import tnames
from pytplot import time_string, time_datetime
from pytplot import time_float, time_double
from .utilities.tcopy import tcopy
from pytplot import tkm2re

from .analysis.avg_data import avg_data
from pytplot import clean_spikes
from .analysis.deriv_data import deriv_data
from .analysis.dpwrspc import dpwrspc
from pytplot import subtract_average
from pytplot import subtract_median
from pytplot import time_clip
from .analysis.tdeflag import tdeflag
from .analysis.tdpwrspc import tdpwrspc
from .analysis.tinterpol import tinterpol
from pytplot import tnormalize
from pytplot import tdotp
from pytplot import tcrossp
from pytplot import tsmooth
from .analysis.yclip import yclip
from .analysis.twavpol import twavpol
from pytplot import cdf_to_tplot

from .cotrans.cotrans import cotrans
from .cotrans.cotrans_get_coord import cotrans_get_coord
from .cotrans.cotrans_set_coord import cotrans_set_coord
from .cotrans.tvector_rotate import tvector_rotate

from .mms import mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_edi, \
    mms_load_edp, mms_load_eis, mms_load_feeps, \
    mms_load_hpca, mms_load_fpi, mms_load_aspoc, \
    mms_load_dsp, mms_load_fsm, mms_load_state
from .mms.feeps.mms_feeps_pad import mms_feeps_pad
from .mms.feeps.mms_feeps_gpd import mms_feeps_gpd
from .mms.eis.mms_eis_pad import mms_eis_pad
from .mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from .mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum

from .maven import maven_load
from .sosmag.load import sosmag_load

from . import erg
from . import ulysses
from . import mica
from . import goes
from . import themis
from . import omni
from . import dscovr
from . import psp
from . import poes
from . import rbsp
from . import ace
from . import wind
from . import csswe
from . import cluster
from . import geotail
from . import twins
from . import stereo
from . import image
from . import polar
from . import fast
from . import equator_s
from . import solo
from . import secs
from . import kyoto
from . import swarm
from . import vires
from . import cnofs
from . import lanl
from . import st5
from . import de2
from . import akebono
from . import soho
from . import barrel
from . import elfin

# set up logging/console output
import logging
from os import environ

logging_level = environ.get('PYTPLOT_LOGGING_LEVEL')
logging_format = environ.get('PYTPLOT_LOGGING_FORMAT')
logging_date_fmt = environ.get('PYTPLOT_LOGGING_DATE_FORMAT')

if logging_format is None:
    logging_format = '%(asctime)s: %(message)s'

if logging_date_fmt is None:
    logging_date_fmt = '%d-%b-%y %H:%M:%S'

if logging_level is None:
    logging_level = logging.INFO
else:
    logging_level = logging_level.lower()
    if logging_level == 'debug':
        logging_level = logging.DEBUG
    elif logging_level == 'info':
        logging_level = logging.INFO
    elif logging_level == 'warn' or logging_level == 'warning':
        logging_level = logging.WARNING
    elif logging_level == 'error':
        logging_level = logging.ERROR
    elif logging_level == 'critical':
        logging_level = logging.CRITICAL

logging.captureWarnings(True)

# basicConfig here doesn't work if it has previously been called
logging.basicConfig(format=logging_format, datefmt=logging_date_fmt, level=logging_level)

# manually set the logger options from the defaults/environment variables
logger = logging.getLogger()
logger_handler = logger.handlers[0]  # should exist since basicConfig has been called
logger_fmt = logging.Formatter(logging_format, logging_date_fmt)
logger_handler.setFormatter(logger_fmt)
logger.setLevel(logging_level)



import os
import re
import warnings
import requests
import logging
import fnmatch
import datetime
import pkg_resources

from pathlib import Path
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from html.parser import HTMLParser
from netCDF4 import Dataset
from cdflib import CDF


# the following is used to parse the links from an HTML index file
class LinkParser(HTMLParser):
    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            attrs = {k: v for (k, v) in attrs}
            if 'href' in attrs:
                link = attrs['href']
                # kludge to support http://rbspice?.ftecs.com/
                if '/' in link:
                    link = link.split('/')[-1]
                try:
                    self.links.append((link))
                except AttributeError:
                    self.links = [(link)]


def check_downloaded_file(filename):
    """
    Check if a file exists and if it can be opened (for CDF and netCDF files).

    If the file exists but it is not CDF or netCDF, it returns True without trying to open the file.
    """
    result = False
    fpath = Path(filename)
    if fpath.is_file() and len(filename) > 3:
        if filename[-4:] == '.cdf':
            # Try to open the cdf file
            try:
                cdf_file = CDF(filename)
                result = True
            except:
                logging.info("Cannot open CDF file: " + filename)
                result = False
        elif filename[-3:] == '.nc':
            # Try to open the netCDF file
            try:
                netcdf_file = Dataset(filename)
                result = True
            except:
                logging.info("Cannot open netCDF file: " + filename)
                result = False
        else:
            # The file is not CDF or netCDF, print a warning and return true
            logging.info("The file is not CDF or netCDF. Filename: " + filename)
            result = True

    return result


def download_file(url=None,
                  filename=None,
                  headers={},
                  username=None,
                  password=None,
                  verify=False,
                  session=None,
                  basic_auth=False,
                  nbr_tries=0):
    """
    Download a file and return its local path; this function is primarily meant to be called by the download function

    Parameters:
        url: str
            Remote URL to download

        filename: str
            Local file name

        headers: dict
            Dictionary containing the headers to be passed to the requests get call

        username: str
            Username to be used in HTTP authentication

        password: str
            password to be used in HTTP authentication

        verify: bool
            Flag indicating whether to verify the SSL/TLS certificate

        session: requests.Session object
            Requests session object that allows you to persist things like HTTP authentication through multiple calls

        nbr_tries: int
            Counts how many times we tried to download the file. Default is 0.

    Notes:
        Checks if the CDF or netCDF file can be opened, and if it can't, tries to download the file for a second time.

    Returns:
        String containing the local file name

    """
    headers_original = headers
    session_original = session

    if session is None:
        session = requests.Session()

    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    # check if the file exists, and if so, set the last modification time in the header
    # this allows you to avoid re-downloading files that haven't changed
    if os.path.exists(filename):
        mod_tm = (datetime.datetime.utcfromtimestamp(os.path.getmtime(filename))).strftime('%a, %d %b %Y %H:%M:%S GMT')
        headers['If-Modified-Since'] = mod_tm

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        if not basic_auth:
            fsrc = session.get(url, stream=True, verify=verify, headers=headers)
        else:
            fsrc = session.get(url, stream=True, verify=verify, headers=headers, auth=(username, password))

    # need to delete the If-Modified-Since header so it's not set in the dictionary in subsequent calls
    if headers.get('If-Modified-Since') is not None:
        del headers['If-Modified-Since']

    needs_to_download_file = False
    if fsrc.status_code == 304:
        # the file hasn't changed
        logging.info('File is current: ' + filename)
        fsrc.close()
    elif fsrc.status_code == 404:
        # file not found
        logging.error('Remote file not found: ' + url)
        fsrc.close()
        return None
    elif fsrc.status_code == 401 or fsrc.status_code == 403:
        # authentication issues
        logging.error('Unauthorized: ' + url)
        fsrc.close()
        return None
    elif fsrc.status_code == 200:
        # this is the main download case
        needs_to_download_file = True
        logging.info('Downloading ' + url + ' to ' + filename)
    else:
        # all other problems
        logging.error(fsrc.reason)
        fsrc.close()
        return None

    if needs_to_download_file:
        ftmp = NamedTemporaryFile(delete=False)

        with open(ftmp.name, 'wb') as f:
            copyfileobj(fsrc.raw, f)

        # make sure the directory exists
        if not os.path.exists(os.path.dirname(filename)) and os.path.dirname(filename) != '':
            os.makedirs(os.path.dirname(filename))

        # if the download was successful, copy to data directory
        copy(ftmp.name, filename)

        fsrc.close()
        ftmp.close()
        os.unlink(ftmp.name)  # delete the temporary file

        logging.info('Download complete: ' + filename)

    # At this point, we check if the file can be opened.
    # If it cannot be opened, we delete the file and try again.
    if nbr_tries == 0 and check_downloaded_file(filename) == False:
        nbr_tries = 1
        logging.info('There was a problem with the file: ' + filename)
        logging.info('We are going to download it for a second time.')
        if os.path.exists(filename):
            os.unlink(filename)

        download_file(url=url,
                      filename=filename,
                      headers=headers_original,
                      username=username,
                      password=password,
                      verify=verify,
                      session=session_original,
                      basic_auth=basic_auth,
                      nbr_tries=nbr_tries)

    # If the file again cannot be opened, we give up.
    if nbr_tries > 0 and check_downloaded_file(filename) == False:
        nbr_tries = 2
        logging.info('Tried twice. There was a problem with the file: ' + filename)
        logging.info('File will be removed. Try to download it again at a later time.')
        if os.path.exists(filename):
            os.unlink(filename)
        filename = None

    return filename


def download(remote_path='',
             remote_file='',
             local_path='',
             local_file='',
             headers={},
             username=None,
             password=None,
             verify=True,
             session=None,
             no_download=False,
             last_version=False,
             basic_auth=False,
             regex=False,
             no_wildcards=False):
    """
    Download one or more remote files and return their local paths.

    Parameters:
        remote_path: str
            String consisting of a common URL base for all remote files

        remote_file: str or list of str
            String or string array of URLs to remote files

        local_path: str
            String consisting of a common local path for all local files

        local_file: str or list of str
            String or string array of local destination file names

        headers: dict
            Dictionary containing the headers to be passed to the requests get call

        username: str
            Username to be used in HTTP authentication

        password: str
            Password to be used in HTTP authentication

        basic_auth: bool
            Flag to indicate that the remote server uses basic authentication
            instead of digest authentication

        verify: bool
            Flag indicating whether to verify the SSL/TLS certificate

        session: requests.Session object
            Requests session object that allows you to persist things like HTTP authentication through multiple calls

        no_download: bool
            Flag to not download remote files

        last_version: bool
            Flag to only download the last in file in a lexically sorted
            list when multiple matches are found using wildcards

        regex: bool
            Flag to allow regular expressions in the file name matching,
            instead of unix style matching

        no_wildcards: bool
            Flag to assume no wild cards in the requested url/filename

    Returns:
        String list specifying the full local path to all requested files

    """
    local_file_in = local_file

    if isinstance(remote_path, list):
        logging.error('Remote path must be a string')
        return

    if isinstance(local_path, list):
        logging.error('Local path must be a string')
        return

    if local_path == '':
        local_path = str(Path('').resolve())

    if username is not None and password is None:
        logging.error('Username provided without password')
        return

    if session is None:
        session = requests.Session()

    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    if headers.get('User-Agent') is None:
        try:
            release_version = pkg_resources.get_distribution("pyspedas").version
        except pkg_resources.DistributionNotFound:
            release_version = 'bleeding edge'
        headers['User-Agent'] = 'pySPEDAS ' + release_version

    out = []
    index_table = {}

    if not isinstance(remote_file, list):
        remote_file = [remote_file]

    urls = [remote_path+rfile for rfile in remote_file]

    for url in urls:
        resp_data = None
        url_file = url[url.rfind("/")+1:]
        url_base = url.replace(url_file, '')

        # automatically use remote_file locally if local_file is not specified
        if local_file_in == '':
            # if remote_file is the entire url then only use the filename
            if remote_path == '':
                local_file = url_file
            else:
                local_file = url.replace(remote_path, '')

                if local_file == '':  # remote_path was the full file name
                    local_file = remote_path[remote_path.rfind("/")+1:]

        filename = os.path.join(local_path, local_file)

        short_path = local_file[:1+local_file.rfind("/")]

        if not no_download:
            # expand the wildcards in the url
            if ('?' in url or '*' in url or regex) and (not no_download and not no_wildcards):
                if index_table.get(url_base) is not None:
                    links = index_table[url_base]
                else:
                    logging.info('Downloading remote index: ' + url_base)

                    # we'll need to parse the HTML index file for the file list
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=ResourceWarning)
                        try:
                            if not basic_auth:
                                html_index = session.get(url_base, verify=verify, headers=headers)
                            else:
                                html_index = session.get(url_base, verify=verify, headers=headers, auth=(username, password))
                        except requests.exceptions.ConnectionError:
                            continue

                    if html_index.status_code == 404:
                        logging.error('Remote index not found: ' + url_base)
                        continue

                    if html_index.status_code == 401 or html_index.status_code == 403:
                        logging.error('Unauthorized: ' + url_base)
                        continue

                    # grab the links
                    link_parser = LinkParser()
                    link_parser.feed(html_index.text)

                    try:
                        links = link_parser.links
                        index_table[url_base] = links
                    except AttributeError:
                        links = []

                # find the file names that match our string
                if not regex:
                    # note: fnmatch.filter accepts ? (single character) and * (multiple characters)
                    new_links = fnmatch.filter(links, url_file)
                else:
                    reg_expression = re.compile(url_file)
                    new_links = list(filter(reg_expression.match, links))

                if len(new_links) == 0:
                    logging.info("No links matching pattern %s found at remote index %s", url_file, url_base)

                if last_version and len(new_links) > 1:
                    new_links = sorted(new_links)
                    new_links = [new_links[-1]]

                if '?' in remote_path or '*' in remote_path:
                    # the user specified a wild card in the remote_path
                    remote_path = url_base

                # download the files
                for new_link in new_links:
                    resp_data = download(remote_path=remote_path, remote_file=short_path+new_link,
                                         local_path=local_path, username=username, password=password,
                                         verify=verify, headers=headers, session=session, basic_auth=basic_auth)
                    if resp_data is not None:
                        for file in resp_data:
                            out.append(file)
                session.close()
                continue

            resp_data = download_file(url=url, filename=filename, username=username, password=password, verify=verify,
                                      headers=headers, session=session, basic_auth=basic_auth)

        if resp_data is not None:
            if not isinstance(resp_data, list):
                resp_data = [resp_data]
            for file in resp_data:
                out.append(file)
        else:
            # download wasn't successful, search for local files
            logging.info('Searching for local files...')
            
            temp_out = []
            

            if local_path == '':
                local_path_to_search = str(Path('.').resolve())
            else:
                local_path_to_search = local_path

            for dirpath, dirnames, filenames in os.walk(local_path_to_search):
                local = local_file[local_file.rfind("/")+1:]
                if not regex:
                    matching_files = fnmatch.filter(filenames, local)
                else:
                    reg_expression = re.compile(local)
                    matching_files = list(filter(reg_expression.match, filenames))

                for file in matching_files:
                    # out.append(os.path.join(dirpath, file))
                    temp_out.append(os.path.join(dirpath, file))
                                
            # check if the file exists, and if so, set the last modification time in the header
            if len(temp_out) == 0:
                logging.info('No local files found for ' + url)
                continue
            temp_out = sorted(temp_out)
            
            if last_version:
                out.append(temp_out[-1]) # append the latest version
            else:
                for file in temp_out:
                    out.append(file)


    session.close()
    return out



"""
Coordinate transformations.

Transform from/to the following coordinate systems:
GSE, GSM, SM, GEI, GEO, MAG, J2000

Times are in Unix seconds for consistency.

Notes
-----
This function is similar to cotrans.pro of IDL SPEDAS.
"""
import logging
import pytplot
from pytplot import get_coords,set_coords
from pyspedas.cotrans.cotrans_lib import subcotrans


def cotrans(name_in=None, name_out=None, time_in=None, data_in=None,
            coord_in=None, coord_out=None):
    """
    Transform data from coord_in to coord_out.

    Parameters
    ----------
    name_in: str, optional
        Pytplot name for input data.
    name_out: str, optional
        Pytplot name for output data.
    time_in: list of float, optional
        Time array.
        Ignored if name_in is provided.
    data_in: list of float, optional
        Data in the coord_in system.
        Ignored if name_in is provided.
    coord_in: str
        Name of input coordinate system.
    coord_out: str
        Name of output coordinate system.

    Returns
    -------
    Returns 1 for successful completion.
        Fills a new pytplot variable with data in the coord_out system.
    """
    if coord_out is None:
        logging.error("cotrans error: No output coordinates were provided.")
        return 0

    # Input data may be specified as a bare array rather than a tplot variable
    if not (name_in is None):
        var_coord_in = get_coords(name_in)
    else:
        var_coord_in = None

    # If the input coordinate system is supplied as an argument, and the tplot variable has a coordinate system
    # specified in its metadata, check that they match, and if not, log the error and return failure.

    if not (var_coord_in is None) and not(coord_in is None):
        if var_coord_in.lower() != coord_in.lower():
            logging.error("cotrans error: " + name_in + " has " +
                          var_coord_in.lower() + " coordinates, but transform from " + coord_in.lower() + " was requested.")
            return 0

    if coord_in is None:
        coord_in = var_coord_in
        if coord_in is None:
            logging.error("cotrans error: No input coordinates were provided.")
            return 0

    coord_in = coord_in.lower()
    coord_out = coord_out.lower()
    all_coords = ["gse", "gsm", "sm", "gei", "geo", "mag", "j2000"]

    if coord_in not in all_coords:
        logging.error("cotrans error: Requested input coordinate system %s not supported.",coord_in)
        return 0
    if coord_out not in all_coords:
        logging.error("cotrans error: Requested output coordinate system %s not supported.",coord_out)
        return 0

    if name_in is not None:
        # If a name_in is provided, use it for data.
        tplot_data = pytplot.get_data(name_in)
        time_in = tplot_data[0]
        data_in = tplot_data[1]
    else:
        pytplot.store_data('cotranstemp', data={'x': list(time_in),
                                                'y': list(data_in)})

    if len(data_in[:]) < 1:
        logging.error("cotrans error: Data is empty.")
        return 0

    # Perform coordinate transformation.
    data_out = subcotrans(list(time_in), list(data_in), coord_in, coord_out)

    if name_in is None and name_out is None:
        return data_out

    if name_in is None:
        name_in = 'cotranstemp'

    # Find the name of the output pytplot variable.
    if name_out is None:
        name_out = name_in + "_" + coord_out

    # Save output pytplot variable.
    pytplot.tplot_copy(name_in, name_out)
    pytplot.data_quants[name_out].data = data_out

    # We should change an attribute for the coordinate system.
    set_coords(name_out, coord_out.upper())

    # Code to update the legend and axis labels has been moved into cotrans_set_coord().

    logging.info("Output variable: " + name_out)

    return 1




import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        mag_files = pyspedas.barrel.sspc(probe='1A', downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))


if __name__ == '__main__':
    unittest.main()



import cdflib
from pytplot import options, ylim, zlim, get_data

from ..load import load


def pwe_ofa(trange=['2017-04-01', '2017-04-02'],
            datatype='spec',
            level='l2',
            suffix='',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            uname=None,
            passwd=None,
            time_clip=False,
            ror=True):
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        ror: bool
            If set, print PI info and rules of the road

    Returns:
        List of tplot variables created.

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24
    prefix = 'erg_pwe_ofa_'+level+'_'+datatype+'_'

    pathformat = 'satellite/erg/pwe/ofa/'+level+'/'+datatype + \
        '/%Y/%m/erg_pwe_ofa_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)

    if (len(loaded_data) > 0) and ror:

        try:
            if isinstance(loaded_data, list):
                if downloadonly:
                    cdf_file = cdflib.CDF(loaded_data[-1])
                    gatt = cdf_file.globalattsget()
                else:
                    gatt = get_data(loaded_data[-1], metadata=True)['CDF']['GATT']
            elif isinstance(loaded_data, dict):
                gatt = loaded_data[list(loaded_data.keys())[-1]]['CDF']['GATT']

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE OFA')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: "+gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/OFA: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Ofa')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    # set spectrogram plot option
    options(prefix+'E_spectra_132'+suffix,  'Spec', 1)
    options(prefix+'B_spectra_132'+suffix,  'Spec', 1)

    # set y axis to logscale
    options(prefix+'E_spectra_132'+suffix,  'ylog', 1)
    options(prefix+'B_spectra_132'+suffix,  'ylog', 1)

    if prefix+'E_spectra_132'+suffix in loaded_data:
        # set ylim
        ylim(prefix+'E_spectra_132'+suffix, 32e-3, 20.)
        # set zlim
        zlim(prefix+'E_spectra_132'+suffix, 1e-9, 1e-2)

    if prefix+'B_spectra_132'+suffix in loaded_data:
        # set ylim
        ylim(prefix+'B_spectra_132'+suffix, 32e-3, 20.)
        # set zlim
        zlim(prefix+'B_spectra_132'+suffix, 1e-4, 1e2)

    # set ytitle
    options(prefix+'E_spectra_132'+suffix,  'ytitle', 'ERG PWE/OFA-SPEC (E)')
    options(prefix+'B_spectra_132'+suffix,  'ytitle', 'ERG PWE/OFA-SPEC (B)')

    # set ysubtitle
    options(prefix+'E_spectra_132'+suffix,  'ysubtitle', 'frequency [kHz]')
    options(prefix+'B_spectra_132'+suffix,  'ysubtitle', 'frequency [kHz]')

    # set ztitle
    options(prefix+'E_spectra_132'+suffix,  'ztitle', 'mV^2/m^2/Hz')
    options(prefix+'B_spectra_132'+suffix,  'ztitle', 'pT^2/Hz')

    # set z axis to logscale
    options(prefix+'E_spectra_132'+suffix,  'zlog', 1)
    options(prefix+'B_spectra_132'+suffix,  'zlog', 1)

    # change colormap option
    options(prefix+'E_spectra_132'+suffix,  'Colormap', 'jet')
    options(prefix+'B_spectra_132'+suffix,  'Colormap', 'jet')

    return loaded_data



import logging
import pyspedas
import pyspedas.themis
from pytplot import data_exists
from pytplot import time_double
from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
import pytplot


def load_needed(trange_loaded,
                trange_needed,
                tolerance=0.0):
    """
    Given a time range of loaded data, and the time range needed, determine if
    a support variable needs to be reloaded.  A small amount of extrapolation is considered
    acceptable, so the time comparisons are done with a user-specified tolerance.

    Parameters:
        trange_loaded: list of float
            Start and end times (Unix times) for support variable currently loaded.

        trange_needed: list of float
            Start and end times (Unix times) for support variable needed to support desired operation.

        tolerance: float
            A duration, in seconds, for which extrapolation from currently loaded data is considered valid.
    """
    st = trange_loaded[0] - tolerance
    et = trange_loaded[1] + tolerance
    if (trange_needed[0] > st) and (trange_needed[1] < et):
        return False
    else:
        return True


def autoload_support(varname=None,
                     trange=None,
                     probe=None,
                     spinaxis: bool = False,
                     spinmodel: bool = False,
                     slp: bool = False):
    """
    Automatically load THEMIS support data required to cover a given probe or time range.

    Parameters:
        varname: str (optional)
            Name of a tplot variable for which calibration, cotrans, or other operation requiring
                support data is needed.  The trange and probe arguments are optional if this argument
                is provided.

        trange: list of Unix times or time strings
            Start and end times (in either string or Unix time format) for which support data is needed.
            Required if varname not specified.

        probe: str (optional)
            A single letter probe identifier.  Required if varname not specified, and spinaxis or spinmodel
                support data is requested.

        spinaxis: bool
            If True, the spin axis variables from the state CDF are examined to see if state needs to
            be reloaded.

        spinmodel: bool
            If True, the spin model produced from the state CDF is examined to see if state needs to
            be reloaded.

        slp: bool
            If True, the tplot variables holding sun and moon positions and lunar coordinate system axes
            are examined to see if the SLP data needs to be reloaded.


    """
    if varname is not None:
        trange = pytplot.get_timespan(varname)
    elif trange is None:
        logging.error("Must specify either a tplot name or a time range in order to load support data")
        return
    elif (probe is None) and (spinaxis or spinmodel):
        logging.error("Must specify either a tplot name or a probe in order to load spin axis or spin model data")
        return

    # Validate varname if present
    if (varname is not None) and not (data_exists(varname)):
        logging.error("tplot variable name " + varname + " not found.")
        return

    # Set probe name (if needed)
    if spinaxis or spinmodel:
        if probe is None:
            probe = varname[2]

    # Set time range (if needed)
    if trange is None:
        trange_needed = pytplot.get_timespan(varname)
    else:
        if isinstance(trange[0], str):
            trange_needed = time_double(trange)
        else:
            trange_needed = trange

    do_state = False
    do_slp = False
    slop = 120.0  # Tolerance (seconds) for determining if existing data covers desired range

    # Does spin model cover desired time interval?
    if spinmodel:
        sm = get_spinmodel(probe, correction_level=1, quiet=True)
        if sm is None:
            do_state = True
        else:
            sminfo = sm.get_timerange()
            trange_loaded = sminfo
            if load_needed(trange_loaded, trange_needed, tolerance=slop):
                do_state = True

    # Do spin axis variables exist, and cover desired time interval?

    if spinaxis:
        v1 = "th" + probe + "_spinras"
        v2 = "th" + probe + "_spindec"
        v3 = "th" + probe + "_spinras_corrected"
        v4 = "th" + probe + "_spindec_corrected"

        # Check uncorrected variables
        if not (data_exists(v1)) or not (data_exists(v2)):
            do_state = True
        else:
            v1_tr = pytplot.get_timespan(v1)
            v2_tr = pytplot.get_timespan(v2)
            if (load_needed(v1_tr, trange_needed, tolerance=slop) or
                    load_needed(v2_tr, trange_needed, tolerance=slop)):
                do_state = True

        # Check corrected variables.  They may be unavailable even if state is reloaded,
        # so only force a reload if one of the variables exists, but doesn't cover the
        # needed time range.

        if data_exists(v3):
            v3_tr = pytplot.get_timespan(v3)
            if load_needed(v3_tr, trange_needed, tolerance=slop):
                do_state = True

        if data_exists(v4):
            v4_tr = pytplot.get_timespan(v4)
            if load_needed(v4_tr, trange_needed, tolerance=slop):
                do_state = True

    # Check SLP variables.  They must all exist, and cover the desired time range, or
    # reload is necessary.

    if slp:
        v1 = 'slp_lun_att_x'
        v2 = 'slp_lun_att_z'
        v3 = 'slp_lun_pos'
        v4 = 'slp_sun_pos'
        if not (data_exists(v1) and data_exists(v2) and data_exists(v3) and data_exists(v4)):
            do_slp = True
        else:
            v1_tr = pytplot.get_timespan(v1)
            v2_tr = pytplot.get_timespan(v2)
            v3_tr = pytplot.get_timespan(v3)
            v4_tr = pytplot.get_timespan(v4)
            if (load_needed(v1_tr, trange_needed, tolerance=slop)
                    or load_needed(v2_tr, trange_needed, tolerance=slop)
                    or load_needed(v3_tr, trange_needed, tolerance=slop)
                    or load_needed(v4_tr, trange_needed, tolerance=slop)):
                do_slp = True

    # Perform the needed updates
    if do_slp:
        pyspedas.themis.slp(trange=trange_needed)
    if do_state:
        pyspedas.themis.state(probe=probe, trange=trange_needed, get_support_data=True)



import unittest
import pyspedas
from pytplot import data_exists
from pyspedas.mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from pyspedas.mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum


class LoadTestCases(unittest.TestCase):
    def test_load_fgm_ql(self):
        data = pyspedas.mms.fgm(probe=1, level='ql', instrument='dfg', trange=['2015-12-15', '2015-12-16']) 
        self.assertTrue(data_exists('mms1_dfg_srvy_dmpa'))
        self.assertTrue(data_exists('mms1_dfg_srvy_gsm_dmpa'))

    def test_load_dfg_l2pre(self):
        data = pyspedas.mms.fgm(probe=1, level='l2pre', instrument='dfg', trange=['2015-12-15', '2015-12-16']) 
        self.assertTrue(data_exists('mms1_dfg_b_gse_srvy_l2pre'))
        self.assertTrue(data_exists('mms1_dfg_b_gsm_srvy_l2pre'))
        self.assertTrue(data_exists('mms1_dfg_b_dmpa_srvy_l2pre'))

    def test_load_afg_l2pre(self):
        data = pyspedas.mms.fgm(probe=1, level='l2pre', instrument='afg', trange=['2015-12-15', '2015-12-16']) 
        self.assertTrue(data_exists('mms1_afg_b_gse_srvy_l2pre'))
        self.assertTrue(data_exists('mms1_afg_b_gsm_srvy_l2pre'))
        self.assertTrue(data_exists('mms1_afg_b_dmpa_srvy_l2pre'))

    def test_load_scm_l1a(self):
        data = pyspedas.mms.scm(probe=1, level='l1a', trange=['2015-12-15', '2015-12-16'])
        self.assertTrue(data_exists('mms1_scm_acb_scm123_scsrvy_srvy_l1a'))

    def test_load_scm_l1a_brst(self):
        data = pyspedas.mms.scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], level='l1a')
        self.assertTrue(data_exists('mms1_scm_acb_scm123_scb_brst_l1a'))
        self.assertTrue(data_exists('mms1_scm_acb_scm123_schb_brst_l1a'))

    def test_load_scm_l1b(self):
        data = pyspedas.mms.scm(probe=4, level='l1b', trange=['2015-12-15', '2015-12-16'])
        self.assertTrue(data_exists('mms4_scm_acb_scm123_scsrvy_srvy_l1b'))

    def test_load_edp_l1b(self):
        data = pyspedas.mms.edp(probe=1, level='l1b', trange=['2015-12-15', '2015-12-16'])
        self.assertTrue(data_exists('mms1_edp_dce_sensor'))
        self.assertTrue(data_exists('mms1_edp_dcv_sensor'))

    def test_load_edi_ql(self):
        data = pyspedas.mms.edi(probe=1, level='ql', trange=['2015-12-15', '2015-12-16'])
        self.assertTrue(data_exists('mms1_edi_E_dmpa'))
        self.assertTrue(data_exists('mms1_edi_v_ExB_dmpa'))

    def test_load_fpi_sitl(self):
        data = pyspedas.mms.fpi(probe=4, level='sitl', trange=['2015-12-15', '2015-12-16']) 
        self.assertTrue(data_exists('mms4_fpi_ePitchAngDist_avg'))

    def test_load_fpi_ql(self):
        data = pyspedas.mms.fpi(probe=1, level='ql', trange=['2015-12-15', '2015-12-16']) 
        self.assertTrue(data_exists('mms1_des_energyspectr_omni_fast'))
        self.assertTrue(data_exists('mms1_des_energyspectr_py_fast'))

    def test_load_hpca_l1b(self):
        trange = ['2015-10-16/13:06', '2015-10-16/13:07']
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='ion')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='combined')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='rf_corr')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='count_rate')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='flux')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='vel_dist')
        data = pyspedas.mms.hpca(probe=1, level='l1b', trange=trange, data_rate='brst', datatype='bkgd_corr')
        mms_hpca_calc_anodes(fov=[0, 360])
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))
        self.assertTrue(data_exists('mms1_hpca_heplus_number_density'))
        self.assertTrue(data_exists('mms1_hpca_heplusplus_number_density'))
        self.assertTrue(data_exists('mms1_hpca_oplus_number_density'))
        self.assertTrue(data_exists('mms1_hpca_heplusplus_RF_corrected'))
        self.assertTrue(data_exists('mms1_hpca_oplusplus_count_rate_elev_0-360'))


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pytplot import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.geotail.mgf(downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_mgf_data(self):
        mgf_vars = pyspedas.geotail.mgf(time_clip=True)
        self.assertTrue(data_exists('IB_vector'))

        mgf_vars = pyspedas.geotail.mgf(datatype='edb3sec', trange=['1998-11-3/09:18:00', '1998-11-3/09:28:00'])
        self.assertTrue(data_exists('BGSE'))

    def test_load_efd_data(self):
        efd_vars = pyspedas.geotail.efd()
        self.assertTrue(data_exists('Es'))

    def test_load_lep_data(self):
        lep_vars = pyspedas.geotail.lep()
        self.assertTrue(data_exists('N0'))

    def test_load_cpi_data(self):
        cpi_vars = pyspedas.geotail.cpi()
        self.assertTrue(data_exists('SW_P_Den'))

    def test_load_epic_data(self):
        epic_vars = pyspedas.geotail.epic()
        self.assertTrue(data_exists('IDiffI_I'))
        epic_vars = pyspedas.geotail.epic(notplot=True)
        self.assertTrue('IDiffI_I' in epic_vars)

    def test_load_pwi_data(self):
        pwi_vars = pyspedas.geotail.pwi()
        self.assertTrue(data_exists('MCAE_AVE'))

if __name__ == '__main__':
    unittest.main()


from copy import deepcopy
import numpy as np
from scipy.ndimage.interpolation import shift
from scipy.constants import c as const_c
from pyspedas import time_double
from pyspedas.particles.spd_slice2d.slice2d_nearest import slice2d_nearest
from pyspedas.particles.spd_slice2d.slice2d_intrange import slice2d_intrange
from pyspedas.particles.spd_slice2d.slice2d_get_support import slice2d_get_support
from pyspedas.particles.spd_slice2d.slice2d_s2c import slice2d_s2c
from pyspedas.particles.spd_slice2d.tplot_average import tplot_average
from pyspedas.mms.particles.moka_mms_clean_data import moka_mms_clean_data


def mms_pad_fpi(dists,
                #disterr,
                time=None,
                window=None,
                center_time=False,
                trange=None,
                samples=None,
                mag_data=None,
                vel_data=None,
                nbin=18,
                da=45.0,
                da2=45.0,
                pr___0=None,
                pr__90=None,
                pr_180=None,
                subtract_bulk=False,
                units='df_cm',
                oclreal=False,
                norm=False):
    """
    Calculate the pitch-angle distribution (angle vs energy plot) and energy spectrum
    in the omni, para, perp, and anti-para directions using MMS FPI data.
    Also returns the one-count-level.

    Parameters
    ------------
        dists : list of dicts
            List of dictionaries containing the FPI particle distribution data.
        time : str or float, optional
            Time at which the pad will be computed.
        window : float, optional
            Length in seconds from TIME over which data will be averaged.
        center_time : bool, optional
            Flag denoting that TIME should be the midpoint for the window instead of the beginning.
        trange : list, optional
            Two-element time range over which data will be averaged.
        samples : int or float, optional
            Number of nearest samples to TIME to average.
        mag_data : str, optional
            Name of tplot variable containing magnetic field data or 3-vector.
            This will be used for pitch-angle calculation and must be in the
            same coordinates as the particle data.
        vel_data : str, optional
            Name of tplot variable containing the bulk velocity data or 3-vector.
            This will be used for pitch-angle calculation and must be in the
            same coordinates as the particle data.
        nbin : int, optional
            Number of bins in the pitch-angle direction.
        da : float, optional
            Pitch angle range for the "para" spectrum, default = 45.0.
        da2 : float, optional
            Pitch angle range for the "anti-para" spectrum, default = 45.0.
        pr___0 : list or tuple, optional
            Pitch angle range for the "para" spectrum, default = [0, 45].
        pr__90 : list or tuple, optional
            Pitch angle range for the "perp" spectrum, default = [45, 135].
        pr_180 : list or tuple, optional
            Pitch angle range for the "anti-para" spectrum, default = [135, 180].
        subtract_bulk : bool, optional
            Set True to subtract bulk velocity from velocity data.
        units : str, optional
            Units for both the pitch-angle-distribution (pad) and energy spectrum.
            Options are 'eflux' [eV/(cm!U2!N s sr eV)] or 'df_km' [s!U3!N / km!U6!N'].
            The default is 'eflux'.
        oclreal : bool, optional
            Set True to return the real one-count level
        norm : bool, optional
            Set True for normalizing the data at each energy bin.

    Returns
    --------
        results : dict
            A dictionary containing the following results:
            - trange : list
                Two-element time range over which data was averaged.
            - egy : array-like
                Energy values in eV.
            - pa : array-like
                Pitch angle values in degrees.
            - data : array-like, shape (Nenergy, Npitch)
                Pitch-angle distribution (angle vs energy plot).
            - datanorm : array-like, shape (Nenergy, Npitch)
                Normalized pitch-angle distribution (angle vs energy plot).
            - numSlices : int
                Number of slices in the data.
            - nbin : int
                Number of pitch angle bins.
            - units : str
                Units of the returned data.
            - subtract_bulk : bool
                True if bulk velocity was subtracted from velocity data.
            - egyrange : list
                Energy range of the data.
            - spec___0 : array-like
                Energy spectrum in the "para" direction.
            - spec__90 : array-like
                Energy spectrum in the "perp" direction.
            - spec_180 : array-like
                Energy spectrum in the "anti-para" direction.
            - spec_omn : array-like
                Energy spectrum in the omni direction.
            - cnts___0 : array-like
                One-count-level data in the "para" direction.
            - cnts__90 : array-like
                One-count-level data in the "perp" direction.
            - cnts_180 : array-like
                One-count-level data in the "anti-para" direction.
            - cnts_omn : array-like
                One-count-level data in the omni direction.
            - oclv___0 : array-like
                One-count-level data in the "para" direction.
            - oclv__90 : array-like
                One-count-level data in the "perp" direction.
            - oclv_180 : array-like
                One-count-level data in the "anti-para" direction.
            - oclv_omn : array-like
                One-count-level data in the omni direction.
            - eror___0 : array-like
                Error data in the "para" direction.
            - eror__90 : array-like
                Error data in the "perp" direction.
            - eror_180 : array-like
                Error data in the "anti-para" direction.
            - eror_omn : array-like
                Error data in the omni direction.
            - vbulk_para : float
                Bulk velocity in the "para" direction.
            - vbulk_perp_abs : float
                Absolute value of bulk velocity in the "perp" direction.
            - vbulk_vxb : float
                Velocity of the cross product of magnetic field and bulk velocity.
            - vbulk_exb : float
                Velocity of the cross product of electric field and magnetic field.
            - bnrm : float
                Average magnetic field magnitude.
            - Vbulk : float
                Bulk velocity magnitude.
            - bfield : float
                Magnetic field magnitude.
            - species : str
                Name of the particle species.
            - pa_azpol : array-like, shape (Npitch, 2)
                Pitch angle values in degrees in azimuthal and polar coordinates.
    """

    dr = np.pi/180.0
    rd = 1.0/dr

    if pr___0 is None:
        pr___0 = [0.0, da]  # para pitch angle range
    if pr__90 is None:
        pr__90 = [90-da2, 90+da2]  # perp pitch angle range
    if pr_180 is None:
        pr_180 = [180-da, 180.0]  # anti-para pitch angle range

    if trange is None:
        if time is None:
            print('Please specifiy a time or time range over which to compute the pad.  For example: ')
            print('  "time=t, window=w" or "trange=tr" or "time=t, samples=n"')
            return
        if window is None and samples is None:
            samples = 1  # use single closest distribution by default
    else:
        # time range already provided
        trange = [np.nanmin(time_double(trange)), np.nanmax(time_double(trange))]

    if time is not None:
        # get the time range if one was specified
        time = time_double(time)

    # get the time range if a time & window were specified instead
    if trange is None and window is not None:
        if center_time:
            trange = [time - window/2.0, time + window/2.0]
        else:
            trange = [time, time + window]

    # if no time range or window was specified then get a time range
    # from the N closest samples to the specified time
    #   (defaults to 1 if SAMPLES is not defined)
    if trange is None:
        trange = slice2d_nearest(dists, time, samples)

    # check that there is data in the trange before proceeding
    times_ind = slice2d_intrange(dists, trange)

    n_samples = len(times_ind)

    if n_samples < 1:
        print('No particle data in the time window; ')
        print('Time samples may be at low cadence; try adjusting the time window.')
        return

    print(str(n_samples) + ' samples in time window')

    nns = np.nanmin(times_ind)
    nne = np.nanmax(times_ind)

    # check support data
    bfield = slice2d_get_support(mag_data, trange)
    vbulk = slice2d_get_support(vel_data, trange)

    if bfield is None:
        print('Magnetic field data needed to calculate pitch-angles')
        return
    if vbulk is None and subtract_bulk:
        print('Velocity data needed to subtract bulk velocity.')
        return

    species = dists[0]['species']
    if units == 'eflux':
        if species == 'i':
            A = 1  # H+
        elif species == 'e':
            A = 1.0/1836.0

        flux_to_df = A**2 * 0.5447*1e6
        cm_to_km = 1e30
        in_u = np.array([2, -1, 0]) # units_in = 'df_km'
        out_u = np.array([0, 0, 0]) # units_out = 'eflux'
        exp = in_u + out_u


    # pitch angle bins
    kmax = nbin
    pamin = 0.0
    pamax = 180.0
    dpa = (pamax-pamin)/kmax  # bin size
    wpa = pamin + np.arange(kmax)*dpa + 0.5*dpa  # bin center values
    pa_bin = np.append(pamin + np.arange(kmax)*dpa, pamax)

    # azimuthal bins
    amax = 16
    azmin = 0.
    azmax = 180.
    daz = (azmax-azmin)/amax
    waz = azmin + np.arange(amax) * daz + 0.5 * daz
    az_bin = np.append(azmin + np.arange(amax) * daz, azmax)

    # polar bins
    pmax = 16
    polmin = 0.
    polmax = 360.
    dpol = (polmax - polmin)/pmax
    wpol = polmin + np.arange(pmax) * dpol + 0.5 * dpol
    pol_bin = np.append(polmin + np.arange(pmax) * dpol, polmax)

    # PA (az x pol)
    pa_azpol = np.zeros((amax, pmax))
    pa_azpol[:, :] = np.nan

    # energy bins
    wegy = dists[0]['energy'][:, 0, 0]
    if subtract_bulk:
        wegy = np.append(2 * wegy[0] - wegy[1], wegy)

    jmax = len(wegy)
    egy_bin = 0.5*(wegy + shift(wegy, -1))
    egy_bin[jmax - 1] = 2. * wegy[jmax - 1] - egy_bin[jmax - 2]
    egy_bin0 = 2. * wegy[0] - egy_bin[0]
    if egy_bin0 < 0:
        egy_bin0 = 0
    egy_bin = np.append(egy_bin0, egy_bin)

    # prep
    pad = np.zeros((jmax, kmax))
    mmax = 4
    f_dat = np.zeros((jmax, mmax)) # Four elements for para, perp, anti-para and omni directions
    f_psd = np.zeros((jmax, mmax))
    f_err = np.zeros((jmax, mmax))
    f_cnt = np.zeros((jmax, mmax))
    count_pad = np.zeros((jmax, kmax))
    count_dat = np.zeros((jmax, mmax))

    # magnetic field and bulk velocity
    bnrm_avg = [0., 0., 0.]
    babs_avg = 0.
    vbulk_avg = [0.0, 0.0, 0.0]
    vbulk_para = 0.0
    vbulk_perp = 0.0
    vbulk_vxb = 0.0
    vbulk_exb = 0.0

    # main loop
    iecl = 0
    iecm = 0

    for n in np.arange(nns, nne+1):
        data = moka_mms_clean_data(dists[n], units=units)

        # magnetic field direction
        tr = [dists[n]['start_time'], dists[n]['end_time']]
        bfield = tplot_average(mag_data, tr, quiet=True)
        babs = np.sqrt(bfield[0]**2 + bfield[1]**2 + bfield[2]**2)
        bnrm = bfield/babs
        bnrm_avg += bnrm
        babs_avg += babs

        # bulk velocity
        if not subtract_bulk:
            vbulk = np.array([0.0, 0.0, 0.0])

        vbpara = bnrm[0]*vbulk[0]+bnrm[1]*vbulk[1]+bnrm[2]*vbulk[2]
        vbperp = vbulk - vbpara
        vbperp_abs = np.sqrt(vbperp[0]**2+vbperp[1]**2+vbperp[2]**2)
        vxb = np.array([-vbulk[1]*bnrm[2]+vbulk[2]*bnrm[1], -vbulk[2]*bnrm[0]+vbulk[0]*bnrm[2], -vbulk[0]*bnrm[1]+vbulk[1]*bnrm[0]])
        vxbabs = np.sqrt(vxb[0]**2+vxb[1]**2+vxb[2]**2)
        vxbnrm = vxb/vxbabs
        exb = np.array([vxb[1]*bnrm[2]-vxb[2]*bnrm[1],vxb[2]*bnrm[0]-vxb[0]*bnrm[2],vxb[0]*bnrm[1]-vxb[1]*bnrm[0]])
        exbabs = np.sqrt(exb[0]**2+exb[1]**2+exb[2]**2)
        exbnrm = exb/exbabs
        vbulk_para += vbpara
        vbulk_perp += vbperp_abs
        vbulk_vxb += vxbnrm[0]*vbulk[0]+vxbnrm[1]*vbulk[1]+vxbnrm[2]*vbulk[2]
        vbulk_exb += exbnrm[0]*vbulk[0]+exbnrm[1]*vbulk[1]+exbnrm[2]*vbulk[2]
        vbulk_avg += vbulk

        # particle velocities & pitch angles

        # spherical to cartesian
        erest = data['mass']*const_c**2/1e6  # convert mass from eV/(km/s)^2 to eV
        vabs = const_c*np.sqrt(1 - 1/((data['energy']/erest + 1)**2))/1000.0
        vdata = slice2d_s2c(vabs, data['theta'], data['phi'])
        vx = vdata[:, 0]
        vy = vdata[:, 1]
        vz = vdata[:, 2]

        if subtract_bulk:
            vx -= vbulk[0]
            vy -= vbulk[1]
            vz -= vbulk[2]

        # pitch angles
        dp = (bnrm[0]*vx + bnrm[1]*vy + bnrm[2]*vz)/np.sqrt(vx**2+vy**2+vz**2)
        dp[dp > 1.0] = 1.0
        dp[dp < -1.0] = -1.0
        pa = rd*np.arccos(dp)

        # Cartesian to spherical
        vnew, theta, phi = cart_to_sphere(vx, vy, vz)
        data['energy'] = erest*(1.0/np.sqrt(1.0-(vnew*1000.0/const_c)**2)-1.0)  # eV
        data['phi'] = phi
        data['theta'] = theta
        data['pa'] = pa

        azang = 90.0 - data['theta']

        imax = len(data['data_dat'])

        for i in np.arange(0, imax):
            # find energy bin
            j = np.nanargmin(np.abs(egy_bin-data['energy'][i]))
            if egy_bin[j] > data['energy'][i]:
                j -= 1
            if j == jmax:
                j -= 1

            # find pitch-angle bin
            k = np.nanargmin(np.abs(pa_bin-data['pa'][i]))
            if pa_bin[k] > data['pa'][i]:
                k -= 1
            if k == kmax:
                k -= 1

            # find azimuthal bin
            a = np.nanargmin(np.abs(az_bin-azang[i]))
            if az_bin[a] > azang[i]:
                a -= 1
            if a == amax:
                a -= 1

            # find polar bin
            p = np.nanargmin(np.abs(pol_bin-data['phi'][i]))
            if pol_bin[p] > data['phi'][i]:
                p -= 1
            if p == pmax:
                p -= 1

            if j >= 0:
                pa_azpol[a, p] = data['pa'][i]

                # find new eflux
                # If shifted to plasma rest-frame, 'eflux' should be re-evaluated
                # from 'psd' because 'eflux' depends on the particle energy. We don't need to
                # worry about this if we want the output in 'psd'.
                newenergy = wegy[j]
                if units == 'eflux':
                    newdat = data['data_psd'][i]*newenergy**exp[0]*(flux_to_df**exp[1]*cm_to_km**exp[2])
                    newpsd = newdat
                    newerr = data['data_err'][i]*newenergy**exp[0]*(flux_to_df**exp[1]*cm_to_km**exp[2])
                else:
                    newdat = data['data_dat'][i]
                    newpsd = data['data_psd'][i]
                    newerr = data['data_err'][i]

                pad[j, k] += newdat
                count_pad[j, k] += 1

                # energy spectrum (para, perp, anti-para)
                m = -1
                if (pr__90[0] <= data['pa'][i]) and (data['pa'][i] <= pr__90[1]):
                    m = 1
                else:
                    if (pr___0[0] <= data['pa'][i]) and (data['pa'][i] <= pr___0[1]):
                        m = 0
                    if (pr_180[0] <= data['pa'][i]) and (data['pa'][i] <= pr_180[1]):
                        m = 2

                if (m >= 0) and (m <= 2):
                    f_dat[j, m] += newdat
                    f_psd[j, m] += newpsd
                    f_err[j, m] += newerr
                    f_cnt[j, m] += data['data_cnt'][i]
                    count_dat[j, m] += 1

                # energy spectrum (omni-direction)
                m = 3
                f_dat[j, m] += newdat
                f_psd[j, m] += newpsd
                f_err[j, m] += newerr
                f_cnt[j, m] += data['data_cnt'][i]
                count_dat[j, m] += 1
            else:
                iecl += 1
        else:
            iecm += imax

    pad /= count_pad
    f_dat /= count_dat
    f_psd /= count_dat
    f_err /= count_dat
    f_cnt /= count_dat

    vbulk_para /= float(n_samples)
    vbulk_perp /= float(n_samples)
    vbulk_vxb /= float(n_samples)
    vbulk_exb /= float(n_samples)
    vbulk_avg /= float(n_samples)

    bnrm_avg /= float(n_samples)
    babs_avg /= float(n_samples)

    pad = np.nan_to_num(pad, nan=0)

    # angle padding
    padnew = np.zeros((jmax, kmax+2))
    padnew[0:jmax, 1:kmax+1] = pad
    padnew[0:jmax, 0] = padnew[0:jmax, 1]
    padnew[0:jmax, kmax + 1] = padnew[0:jmax, kmax]
    pad = padnew
    wpa_new = np.append(wpa[0] - dpa, wpa)
    wpa_new = np.append(wpa_new, wpa[kmax - 1] + dpa)
    wpa = wpa_new

    # normalize
    padnorm = deepcopy(pad)
    for j in range(0, jmax):
        peak = np.nanmax(pad[j, 0:kmax])  # find the peak
        if peak == 0:
            padnorm[j, 0:kmax] = 0.0
        else:
            padnorm[j, 0:kmax] /= peak

    if norm:
        pad = padnorm

    # Effective one-count-level
    # 'f_psd' is the PSD    averaged over time and angular ranges.
    # 'f_cnt' is the counts averaged over time and angular ranges.
    # 'count_dat is the total number of time and angular bins.
    if oclreal:
        f_ocl = f_psd/f_cnt
    else:
        f_ocl = f_psd/(f_cnt*count_dat)

    # output
    return {'trange': trange,
            'egy': wegy,
            'pa': wpa,
            'data': pad,
            'datanorm': padnorm,
            'numSlices': n_samples,
            'nbin': kmax,
            'units': units,
            'subtract_bulk': subtract_bulk,
            'egyrange': [np.nanmin(wegy), np.nanmax(wegy)],
            #'parange': [np.nanmin(wpa), np.nanmax(wpa)],
            'spec___0': f_psd[:, 0],
            'spec__90': f_psd[:, 1],
            'spec_180': f_psd[:, 2],
            'spec_omn': f_psd[:, 3],
            'cnts___0': f_cnt[:, 0],
            'cnts__90': f_cnt[:, 1],
            'cnts_180': f_cnt[:, 2],
            'cnts_omn': f_cnt[:, 3],
            'oclv___0': f_ocl[:, 0],
            'oclv__90': f_ocl[:, 1],
            'oclv_180': f_ocl[:, 2],
            'oclv_omn': f_ocl[:, 3],
            'eror___0': f_err[:, 0],
            'eror__90': f_err[:, 1],
            'eror_180': f_err[:, 2],
            'eror_omn': f_err[:, 3],
            'vbulk_para': vbulk_para,
            'vbulk_perp_abs': vbulk_perp,
            'vbulk_vxb': vbulk_vxb,
            'vbulk_exb': vbulk_exb,
            'bnrm': bnrm_avg,
            'Vbulk': vbulk_avg,
            'bfield': bnrm_avg * babs_avg,
            'species': species,
            'pa_azpol': pa_azpol,
            'wpol': wpol,
            'waz': waz}


def cart_to_sphere(x, y, z):
    rho = x * x + y * y
    r = np.sqrt(rho + z * z)
    phi = 18e1/np.pi * np.arctan2(y, x)
    # should be between 0-360
    phi[phi < 0] += 360.0
    theta = 18e1/np.pi * np.arctan(z / np.sqrt(rho))
    return (r, theta, phi)


"""Transform GSE data to SSE data.

Notes:
    Works in a similar way to IDL spedas dsl2gse.pro
"""

import logging
import numpy as np
from copy import deepcopy

from pyspedas import tnormalize, tcrossp, tinterpol, deriv_data, cotrans, tvector_rotate
from pyspedas.themis import autoload_support
from pytplot import data_exists, get_data, store_data, get_coords, set_coords

def gse2sse(name_in: str, name_out: str, isssetogse: bool = False,
            variable_type: str = 'Other', ignore_input_coord: bool = False, rotation_only: bool = False) -> int:

    """Transform gse to sse.

    Parameters
    ----------
        name_in: str
            Name of input pytplot variable (e.g. 'tha_fgl_dsl')
        name_out: str
            Name of output pytplot variable (e.g. 'tha_fgl_sse')
        isssetogse: bool
            If False (default), then GSE to SSE.
            If True, then SSE to GSE.
        ignore_input_coord: bool
            if False (default), do not check the input coordinate system
            if True, fail and return 0 if input coordinate does not match the requested transform.
        variable_type: str
            A string describing the type of data being transformed.  If value is "pos" or "vel", the appropriate
                offsets (lunar position or velocity) are applied during the transform.  Any other value will be treated
                as equivalent to rotate_only=True.
        rotation_only: bool
            if False (default), assume input variable is a position with units of km, and apply the earth-moon
                offset before rotating to SSE, or after rotating to GSE
            if True, assume the input variable is a velocity or some other quantity that does not need the earth-moon
                translation step

    Returns
    -------
        1 for sucessful completion.
    """

    needed_vars = [name_in]
    c = [value for value in needed_vars if data_exists(value)]
    if len(c) < 1:
        logging.error("Variables needed: " + str(needed_vars))
        m = [value for value in needed_vars if value not in c]
        logging.error("Variables missing: " + str(m))
        logging.error("Please load missing variables.")
        return 0

    # load support data
    autoload_support(varname=name_in, slp=True)
    name_sun_pos = 'slp_sun_pos'
    name_lun_pos = 'slp_lun_pos'

    if not ignore_input_coord:
        # check input coord
        in_coord = get_coords(name_in)
        if in_coord is None:
            in_coord = "None"
        if not isssetogse and (in_coord.lower() != 'gse'):
            logging.error("GSE to SSE transform requested, but input coordinate system is " + in_coord)
            return 0
        if isssetogse and (in_coord.lower() != 'sse'):
            logging.error("SSE to GSE transform requested, but input coordinate system is " + in_coord)
            return 0
        
        # check sun pos coord
        sun_pos_coord = get_coords(name_sun_pos)
        if sun_pos_coord is None:
            sun_pos_coord = "None"
        if sun_pos_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE',name_sun_pos)
            sun_pos_gse_name = name_sun_pos+'_gse'
            cotrans(name_in=name_sun_pos,name_out=sun_pos_gse_name,coord_out='GSE')
        else:
            sun_pos_gse_name=name_sun_pos

        # check lun pos coord
        lun_pos_coord = get_coords(name_lun_pos)
        if lun_pos_coord is None:
            lun_pos_coord = "None"
        if lun_pos_coord.lower() != 'gse':
            logging.info('Transforming %s to GSE', name_lun_pos)
            lun_pos_gse_name = name_lun_pos + '_gse'
            cotrans(name_in=name_lun_pos, name_out=lun_pos_gse_name, coord_out='GSE')
        else:
            lun_pos_gse_name = name_lun_pos
    else:
        sun_pos_gse_name=name_sun_pos
        lun_pos_gse_name=name_lun_pos

    meta_copy = deepcopy(get_data(name_in, metadata=True))

    # Make rotation matrix
    sunpos = get_data(sun_pos_gse_name)
    lunpos = get_data(lun_pos_gse_name)
    sun_pos_dim = sunpos.y.shape

    # Moon to sun vector = sunpos - lunpos
    lun_sun_vec = sunpos.y - lunpos.y

    # SSE X-axis: unit vector from sun toward moon
    sse_x = tnormalize(lun_sun_vec, return_data=True)
    # SSE Y-axis: ecliptic north unit vector (0,0,1) cross SSE-X
    ecliptic_north = np.repeat(np.array([[0, 0, 1]]), sun_pos_dim[0], axis=0)
    sse_y = tcrossp(ecliptic_north,sse_x, return_data=True)
    # SSE Z-axis: SSE-X cross SSE-Y (not necessarily ecliptic north)
    sse_z = tcrossp(sse_x, sse_y, return_data=True)

    # Make rotation matrix from basis vectors, store in tplot variable

    out_data = np.zeros((sun_pos_dim[0], 3, 3))
    if not isssetogse:
        out_data[:,0,:] = sse_x
        out_data[:,1,:] = sse_y
        out_data[:,2,:] = sse_z
    else:
        # Invert sense of conversion by transposing rotation array
        out_data[:,:,0] = sse_x
        out_data[:,:,1] = sse_y
        out_data[:,:,2] = sse_z

    store_data('sse_mat_cotrans', data={'x': sunpos.times, 'y': out_data})

    if variable_type.lower() == "pos" and not rotation_only:
        tinterpol(lun_pos_gse_name,name_in,newname='gse2sse_offset')
        gse2sse_offset_data = get_data('gse2sse_offset')
    elif variable_type.lower() == "vel" and not rotation_only:
        deriv_data(lun_pos_gse_name,new_names='gse2sse_lun_vel')
        tinterpol('gse2sse_lun_vel',name_in,newname='gse2sse_offset')
        gse2sse_offset_data = get_data('gse2sse_offset')
    else:
        logging.info("No offsets performed for variable type %s",variable_type)
        rotation_only = True

    if not isssetogse:
        """ GSE -> SSE
        """

        if not rotation_only:
            logging.info("Applying earth-moon %s offset to input variable %s",variable_type,name_in)
            input_pos = get_data(name_in)
            translated_pos = input_pos.y - gse2sse_offset_data.y
            name_trans = name_in + '_trans'
            store_data(name_trans,data={'x':input_pos.times, 'y':translated_pos},attr_dict=meta_copy)
            tvector_rotate('sse_mat_cotrans',name_trans,newname=name_out)
        else:
            logging.info("No earth-moon offsets applied")
            tvector_rotate('sse_mat_cotrans',name_in,newname=name_out)

        set_coords(name_out,'SSE')
        return 1
    
    else:
        """ SSE -> GSE
        """
        if not rotation_only:
            tvector_rotate('sse_mat_cotrans',name_in,newname='gse2sse_rotated')
            logging.info("Applying moon-earth %s offset to rotated variable %s",variable_type,'gse2sse_rotated')
            rotated_data = get_data('gse2sse_rotated')
            earth_data = rotated_data.y + gse2sse_offset_data.y
            store_data(name_out,data={'x':rotated_data.times,'y':earth_data},attr_dict=meta_copy)
        else:
            logging.info("No earth-moon offsets applied")
            tvector_rotate('sse_mat_cotrans',name_in,newname=name_out)

        set_coords(name_out,'GSE')
        return 1





""" Tests of spinmodel construction and interpolation """
import unittest
from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose
from pytplot import get_data, store_data, time_string, del_data, cdf_to_tplot
from pyspedas.themis import state, get_spinmodel



class SpinmodelDataValidation(unittest.TestCase):
    """
    Compare spin models and interpolation results generated by IDL and Python.

    The corresponding IDL script takes a time range, probe, and eclipse correction level,
    then creates a tplot variable containing the test parameters.  IDL builds a spin model by calling
    thm_load_state for the given probe and time interval, and dumps all the segment parameters to tplot
    variables.  Then a spin model interpolation routine is called for a set of timestamps, and all the
    interpolated parameters are dumped to tplot variables.   The tplot outputs are saved in a cdf file or a
    tplot save file.

    On the python side, the validation file is read to get the test parameters.  The spin models are
    created with a call to themis.state(), then the test timestamps are passed to the spinmodel interpolation
    routine.   The Python and IDL tplot variables are each given their own prefix.

    After loading the necessary data, the various tests in this file compare the Python and IDL values
    for each of the spinmodel segment parameters, and interpolation results.
    """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates data file: projects/themis/spin/spinmodel_python_test.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Download tplot files
        remote_server = 'https://spedas.org/'
        # remote_name = 'testfiles/thm_cotrans_validate.cdf'
        remote_name = 'testfiles/tha_validate_spinmodel.cdf'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        cdf_to_tplot(filename)
        # pytplot.tplot_restore(filename)
        t_dummy, trange = get_data('parm_trange')
        t_dummy, probe_idx = get_data('parm_probe')
        t_dummy, correction_level = get_data('parm_correction_level')
        #print(trange)
        #print(time_string(trange))
        #print(probe_idx)
        int_idx = int(probe_idx[0])
        probes = ['a', 'b', 'c', 'd', 'e']
        int_corr_level = int(correction_level[0])
        probe = probes[int_idx]
        #print(probe)
        #print(int_corr_level)
        thm_data = state(trange=trange, probe=probe, get_support_data=True)
        cls.model = get_spinmodel(probe, int_corr_level)
        cls.model.make_tplot_vars('py_seg_')
        #pytplot.tplot_names()
        dummy_t, tst_times = get_data('interp_times')
        res = cls.model.interp_t(tst_times)
        store_data('py_spinphase', data={'x': tst_times, 'y': res.spinphase})
        store_data('py_spinper', data={'x': tst_times, 'y': res.spinper})
        store_data('py_spincount', data={'x': tst_times, 'y': res.spincount})
        store_data('py_t_last', data={'x': tst_times, 'y': res.t_last})
        store_data('py_eclipse_delta_phi', data={'x': tst_times, 'y': res.eclipse_delta_phi})
        store_data('py_segflags', data={'x': tst_times, 'y': res.segflags})

    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # pytplot.del_data('*')

    def test_seg_t1(self):
        pydata = get_data('py_seg_t1')
        idldata = get_data('seg_t1')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_t2(self):
        pydata = get_data('py_seg_t2')
        idldata = get_data('seg_t2')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_c1(self):
        pydata = get_data('py_seg_c1')
        idldata = get_data('seg_c1')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_c2(self):
        pydata = get_data('py_seg_c2')
        idldata = get_data('seg_c2')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_b(self):
        pydata = get_data('py_seg_b')
        idldata = get_data('seg_b')
        assert_allclose(pydata.y, idldata.y, rtol=1.0e-05)

    def test_seg_c(self):
        pydata = get_data('py_seg_c')
        idldata = get_data('seg_c')
        assert_allclose(pydata.y, idldata.y, rtol=1.0e-06)

    def test_seg_npts(self):
        pydata = get_data('py_seg_npts')
        idldata = get_data('seg_npts')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_maxgap(self):
        pydata = get_data('py_seg_maxgap')
        idldata = get_data('seg_maxgap')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_phaserr(self):
        pydata = get_data('py_seg_phaserr')
        idldata = get_data('seg_phaserr')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_idpu_spinper(self):
        pydata = get_data('py_seg_idpu_spinper')
        idldata = get_data('seg_idpu_spinper')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_initial_delta_phi(self):
        pydata = get_data('py_seg_initial_delta_phi')
        idldata = get_data('seg_initial_delta_phi')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_seg_segflags(self):
        pydata = get_data('py_seg_segflags')
        idldata = get_data('seg_segflags')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_interp_spinphase(self):
        pydata = get_data('py_spinphase')
        idldata = get_data('interp_spinphase')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-04)

    def test_interp_spinper(self):
        pydata = get_data('py_spinper')
        idldata = get_data('interp_spinper')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_interp_spincount(self):
        pydata = get_data('py_spincount')
        idldata = get_data('interp_spincount')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_interp_segflags(self):
        pydata = get_data('py_segflags')
        idldata = get_data('interp_segflags')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_interp_eclipse_delta_phi(self):
        pydata = get_data('py_eclipse_delta_phi')
        idldata = get_data('interp_eclipse_delta_phi')
        assert_allclose(pydata.y, idldata.y, atol=1.0e-06)

    def test_timerange(self):
        trange=self.model.get_timerange()
        print(time_string(trange[0]),time_string(trange[1]))

    def test_eclipse_times(self):
        start_times,end_times=self.model.get_eclipse_times()
        for i in range(len(start_times)):
            print(time_string(start_times[i]),time_string(end_times[i]))

if __name__ == '__main__':
    unittest.main()



"""
Subtracts the average (mean) or the median from the data.

Notes
-----
Similar to tsub_average.pro in IDL SPEDAS.

"""
import logging
import pytplot


def subtract_average(names, new_names=None, suffix=None, overwrite=None,
                     median=None):
    """
    Subtracts the average or the median from data.

    Parameters
    ----------
    names: str/list of str
        List of pytplot names.
    new_names: str/list of str, optional
        List of new_names for pytplot variables.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-d'.
    overwrite: bool, optional
        If set, then pytplot variables are replaced.
    median: float, optional
        If it is 0 or not set, then it computes the mean.
        Otherwise, it computes the median.

    Returns
    -------
    None.

    """
    logging.info("subtract_average has been moved to the pytplot.tplot_math package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    pytplot.tplot_math.subtract_average(names=names,new_names=new_names,suffix=suffix,overwrite=overwrite,median=median)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot
from .rfs import rfs_variables_to_load
from .config import CONFIG

def load(trange=['2018-11-5', '2018-11-6'],
         instrument='fields',
         datatype='mag_RTN',
         spec_types=None,  # for DFB AC spectral data
         level='l2',
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         username=None,
         password=None,
         last_version=False):
    """
    This function loads Parker Solar Probe data into tplot variables; this function is not 
    meant to be called directly; instead, see the wrappers: 
        psp.fields: FIELDS data
        psp.spc: Solar Probe Cup data
        psp.spe: SWEAP/SPAN-e data
        psp.spi: SWEAP/SPAN-i data
        psp.epihi: ISoIS/EPI-Hi data
        psp.epilo: ISoIS/EPI-Lo data
        psp.epi ISoIS/EPI (merged Hi-Lo) data
    
    """
    # remote path formats generally are going to be all lowercase except for
    # on the Berkeley FIELDS server
    if (username is not None) and (datatype in ['mag_RTN_1min',
                                            'mag_RTN_4_Sa_per_Cyc',
                                            'mag_RTN',
                                            'mag_SC',
                                            'mag_SC_1min',
                                            'mag_SC_4_Sa_per_Cyc',
                                            'sqtn_rfs_V1V2'
                                            ]):
        pass
    else:
        datatype = datatype.lower()

    prefix = 'psp_'  #To cover the case if one *does* call this routine directly.

    file_resolution = 24*3600.
    if instrument == 'fields':
        prefix = '' #CDF Variables are already prefixed with psp_fld_

        # 4_per_cycle and 1min are daily, not 6h like the full resolution 'mag_(rtn|sc)'
        if datatype in ['mag_rtn', 'mag_sc']:
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_v??.cdf'
            file_resolution = 6*3600.
        elif datatype in ['mag_rtn_1min', 'mag_sc_1min', 'rfs_hfr', 'rfs_lfr', 'rfs_burst', 'f2_100bps', 'aeb']:
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v??.cdf'
        elif datatype in ['mag_rtn_4_per_cycle', 'mag_rtn_4_sa_per_cyc']:
            pathformat = instrument + '/' + level + '/mag_rtn_4_per_cycle/%Y/psp_fld_' + level + '_mag_rtn_4_sa_per_cyc_%Y%m%d_v??.cdf'
        elif datatype in ['mag_sc_4_per_cycle', 'mag_sc_4_sa_per_cyc']:
            pathformat = instrument + '/' + level + '/mag_sc_4_per_cycle/%Y/psp_fld_' + level + '_mag_sc_4_sa_per_cyc_%Y%m%d_v??.cdf'
        elif datatype == 'sqtn_rfs_v1v2':
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v?.?.cdf'        
        elif datatype == 'rfs_lfr_qtn':
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m*_v??.cdf'
        elif datatype in ['dfb_dc_spec', 'dfb_ac_spec', 'dfb_dc_xspec', 'dfb_ac_xspec']:
            out_vars = []
            for item in spec_types:
                loaded_data = load(trange=trange, instrument=instrument, datatype=datatype + '_' + item, level=level, 
                    suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
                    downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, last_version=last_version)
                if loaded_data != []:
                    out_vars.extend(loaded_data)
            return out_vars
        elif datatype[:12] == 'dfb_dc_spec_' or datatype[:12] == 'dfb_ac_spec_' or datatype[:13] == 'dfb_dc_xspec_' or datatype[:13] == 'dfb_ac_xspec_':
            if datatype[:13] == 'dfb_dc_xspec_' or datatype[:13] == 'dfb_ac_xspec_':
                dtype_tmp = datatype[:12]
                stype_tmp = datatype[13:]
            else:
                dtype_tmp = datatype[:11]
                stype_tmp = datatype[12:]
            pathformat = instrument + '/' + level + '/' + dtype_tmp + '/' + stype_tmp + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v??.cdf'



        # unpublished data
        elif username != None:
            if datatype in ['mag_RTN', 'mag_SC']:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_v??.cdf'
                file_resolution = 6*3600.

            elif datatype in ['mag_RTN_1min', 'mag_RTN_4_Sa_per_Cyc', 'mag_SC_1min', 'mag_SC_4_Sa_per_Cyc']:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v??.cdf'

            elif datatype ==  'sqtn_rfs_V1V2':
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v?.?.cdf'

            else:
                pathformat = instrument + '/' + level + '/' + datatype + '/%Y/%m/psp_fld_' + level + '_' + datatype + '_%Y%m%d_v??.cdf'

        else:
            # Generic SPDF path.  
            pathformat = instrument + '/' + level + '/' + datatype + '/%Y/psp_fld_' + level + '_' + datatype + '_%Y%m%d%H_v??.cdf'
            file_resolution = 6*3600.

        # Files on Berkeley server are stored in monthly directories 
        if username != None:
            pathformat = pathformat.replace('/%Y/psp_fld', '/%Y/%m/psp_fld')
            if level == 'l1':
                pathformat = pathformat.replace('psp_fld', 'spp_fld')

    elif instrument == 'spc':
        if username is None:
            prefix = 'psp_spc_'
            pathformat = 'sweap/spc/' + level + '/' + datatype + '/%Y/psp_swp_spc_' + datatype + '_%Y%m%d_v??.cdf'
        else:
            # unpublished data
            # spc pre-public data is prepended by "spp", not "psp"
            # The psp_ files are located in the same directory after public release
            # but in that case users can just use the public (spdf) option. 
            prefix = 'spp_spc_'
            pathformat = 'sweap/spc/' + level + '/%Y/%m/spp_swp_spc_' + datatype + '_%Y%m%d_v0?.cdf'
    elif instrument == 'spe':
        prefix = 'psp_spe_'
        pathformat = 'sweap/spe/' + level + '/' + datatype + '/%Y/psp_swp_sp?_*_%Y%m%d_v??.cdf'
    elif instrument == 'spi':
        if username is None:
            prefix = 'psp_spi_'
            pathformat = 'sweap/spi/' + level + '/' + datatype + '/%Y/psp_swp_spi_*_%Y%m%d_v??.cdf'
        else:
            # unpublished data
            prefix = 'psp_spi_'
            pathformat = 'sweap/spi/' + level + '/' + datatype + '/%Y/%m/psp_swp_' + datatype + '*_%Y%m%d_v0?.cdf'
    elif instrument == 'epihi':
        prefix = 'psp_epihi_'
        pathformat = 'isois/epihi/' + level + '/' + datatype + '/%Y/psp_isois-epihi_' + level + '*_%Y%m%d_v??.cdf'
    elif instrument == 'epilo':
        prefix = 'psp_epilo_'
        pathformat = 'isois/epilo/' + level + '/' + datatype + '/%Y/psp_isois-epilo_' + level + '*_%Y%m%d_v??.cdf'
    elif instrument == 'epi':
        prefix = 'psp_isois_'
        pathformat = 'isois/merged/' + level + '/' + datatype + '/%Y/psp_isois_' + level + '-' + datatype + '_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange, res=file_resolution)

    out_files = []

    if username is None:
        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                        local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version)
    else:
        if instrument == 'fields':
            try:
                print("Downloading unpublished Data....")
                files = download(
                    remote_file=remote_names, remote_path=CONFIG['fields_remote_data_dir'], 
                    local_path=CONFIG['local_data_dir'], no_download=no_update,
                    username=username, password=password, basic_auth=True,last_version=last_version
                )
            except:
                files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                                local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version)
        elif instrument in ['spc','spi']:
            try:
                print("Downloading unpublished Data....")
                files = download(
                    remote_file=remote_names, remote_path=CONFIG['sweap_remote_data_dir'], 
                    local_path=CONFIG['local_data_dir'], no_download=no_update,
                    username=username, password=password, basic_auth=True,last_version=last_version
                )
            except:
                files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], 
                                local_path=CONFIG['local_data_dir'], no_download=no_update,last_version=last_version)
        

    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    # find the list of varnames for RFS data
    # these files have > 1500 variables, but
    # we only load ~50
    if 'rfs' in datatype.lower() and varformat is None and varnames == []:
        varnames = rfs_variables_to_load(out_files)
        # we'll need the support data for the quality flags
        get_support_data = True

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, 
                        varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import logging
import numpy as np
import pyspedas 
from pytplot import get, store, options
from pyspedas.mms.feeps.mms_feeps_active_eyes import mms_feeps_active_eyes
from pyspedas.mms.feeps.mms_feeps_getgyrophase import mms_feeps_getgyrophase

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean


def mms_feeps_gpd(trange=['2017-07-11/22:30', '2017-07-11/22:35'], 
                  probe='2', 
                  data_rate='brst', 
                  level='l2', 
                  datatype='electron',
                  data_units='intensity',
                  bin_size=15, # deg
                  energy=[50, 500]):
    """
    Calculate gyrophase distributions using data from the MMS Fly's Eye Energetic Particle Sensor (FEEPS)

    Parameters
    ----------
        probe: str
            probe #, e.g., '4' for MMS4

        data_units: str
            'intensity' 

        datatype: str
            'electron' or 'ion'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level

        suffix: str
            suffix of the loaded data

        energy: list of float
            energy range to include in the calculation

        bin_size: float
            size of the pitch angle bins

    Returns
    --------
        Tplot variable containing the gyrophase distribution

    Notes
    ------
        Based on IDL code by Drew Turner (10 Oct 2017): mms_feeps_gpd.pro
    """

    if isinstance(probe, int):
        probe = str(probe)

    feeps_data = pyspedas.mms.feeps(trange=trange, data_rate=data_rate, probe=probe, level=level)

    if len(feeps_data) == 0:
        logging.error('Problem loading FEEPS data for this time range.')
        return

    # Account for angular response (finite field of view) of instruments
    # elec can use +/-21.4 deg on each pitch angle as average response angle; ions can start with +/-10 deg, but both need to be further refined
    if datatype == 'electron': dAngResp = 21.4 # [deg] 
    if datatype == 'ion': dAngResp = 10.0 # [deg]

    bin_size = float(bin_size)
    n_bins = 360.0/bin_size
    gyro_bins = 360.*np.arange(n_bins+1)/n_bins
    gyro_centers = 360.*np.arange(n_bins)/n_bins+(gyro_bins[1]-gyro_bins[0])/2.

    # get the gyrophase angles
    # calculate the gyro phase angles from the magnetic field data
    gyro_vars = mms_feeps_getgyrophase(trange=trange, probe=probe, data_rate=data_rate, level=level, datatype=datatype)
    gyro_data = get('mms' + str(probe) + '_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_gyrophase')

    if gyro_data is None or gyro_vars is None:
        logging.error('Problem calculating gyrophase angles.')
        return

    eyes = mms_feeps_active_eyes(trange, probe, data_rate, datatype, level)

    data_map = {}

    if data_rate == 'srvy':
        # From Allison Jaynes @ LASP: The 6,7,8 sensors (out of 12) are ions,
        # so in the pitch angle array, the 5,6,7 columns (counting from zero) will be the ion pitch angles.
        # for electrons:
        if datatype == 'electron': 
            data_map['top-electron'] = eyes['top']-1
            data_map['bottom-electron'] = eyes['bottom']-1
        elif datatype == 'ion':
            data_map['top-ion'] = eyes['top']-1
            data_map['bottom-ion'] = eyes['bottom']-1
    elif data_rate == 'brst':
        # note: the following are indices of the top/bottom sensors in pa_data
        # they should be consistent with pa_dlimits.labels
        data_map['top-electron'] = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        data_map['bottom-electron'] = [9, 10, 11, 12, 13, 14, 15, 16, 17]
        # and ions:
        data_map['top-ion'] = [0, 1, 2]
        data_map['bottom-ion'] = [3, 4, 5]

    sensor_types = ['top', 'bottom']

    # First, initialize arrays for flux (dflux) and pitch angles (dpa) compiled from all sensors:
    if datatype == 'electron':
        dflux = np.zeros((len(gyro_data.times), len(data_map['top-electron']) + len(data_map['bottom-electron'])))
    elif datatype == 'ion':
        dflux = np.zeros((len(gyro_data.times), len(data_map['top-ion']) + len(data_map['bottom-ion'])))

    dpa = np.zeros(dflux.shape)

    for sensor_type in sensor_types:
        pa_map = data_map[sensor_type + '-' + datatype]
        particle_idxs = np.array(eyes[sensor_type])-1

        for isen in range(len(particle_idxs)): # loop through sensors
            # get the data
            var_name = 'mms' + str(probe) + '_epd_feeps_' + data_rate + '_' + level + '_' + datatype + '_' + sensor_type + '_' + data_units + '_sensorid_' + str(particle_idxs[isen]+1) + '_clean_sun_removed'
            data = get(var_name)
            if data is None:
                logging.error('Data not found: ' + var_name)
                continue
            data.y[data.y == 0.0] = np.nan # remove any 0s before averaging
            # Energy indices to use:
            indx = np.argwhere((data.v <= energy[1]) & (data.v >= energy[0]))
            if len(indx) == 0:
                logging.error('Energy range selected is not covered by the detector for FEEPS ' + datatype + ' data')
                continue
            dflux[:, pa_map[isen]] = nanmean(data.y[:, indx], axis=1).flatten()
            dpa[:, pa_map[isen]] = gyro_data.y[:,  pa_map[isen]].flatten()

    # we need to replace the 0.0s left in after populating dpa with NaNs; these 
    # 0.0s are left in there because these points aren't covered by sensors loaded
    # for this datatype/data_rate
    dpa[dpa == 0.0] = np.nan

    gyro_flux = np.zeros((len(gyro_data.times), int(n_bins)))
    delta_gyro = (gyro_bins[1]-gyro_bins[0])/2.0

    # Now loop through PA bins and time, find the telescopes where there is data in those bins and average it up!
    for it in range(len(dpa[:, 0])):
        for ipa in range(int(n_bins)):
            ind = np.argwhere((dpa[it, :] + dAngResp >= gyro_centers[ipa]-delta_gyro) & (dpa[it, :] - dAngResp < gyro_centers[ipa]+delta_gyro)).flatten()
            if ind.size != 0:
                if len(ind) > 1:
                    gyro_flux[it, ipa] = nanmean(dflux[it, ind])
                else:
                    gyro_flux[it, ipa] = dflux[it, ind[0]]
            #if len(ind) > 0:
            #   gyro_flux[it, ipa] = np.nanmean(dflux[it, ind], axis=0).flatten()

    # fill any missed bins with NAN
    gyro_flux[gyro_flux == 0.0] = np.nan 

    en_range_string = str(int(energy[0])) + '-' + str(int(energy[1])) + 'keV'

    new_name = 'mms' + str(probe) + '_epd_feeps_' + data_rate + '_' + level + '_' + datatype + '_' + data_units + '_' + en_range_string + '_gpd'

    saved = store(new_name, data={'x': gyro_data.times, 'y': gyro_flux, 'v': gyro_centers})

    if saved:
        options(new_name, 'spec', True)
        options(new_name, 'zlog', False)
        return new_name



"""
This module perform unitest on elfin state file by comparing 
with tplot variable genrate by IDL routine

How to run:
    $ python -m pyspedas.elfin.tests.test_state
"""
import unittest
import logging
import pytplot.get_data
from pytplot.importers.tplot_restore import tplot_restore
from numpy.testing import assert_allclose, assert_array_almost_equal, assert_array_equal

import pyspedas.elfin
from pyspedas.utilities.download import download
from pyspedas.elfin.config import CONFIG

TEST_DATASET_PATH="test/"

class TestELFStateValidation(unittest.TestCase):
    """Tests of the data been identical to SPEDAS (IDL)."""

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The IDL script that creates data file: (epd_state_validation.pro)
        """
        # Testing time range
        cls.t = ['2021-10-12/23:00:00','2021-10-12/23:10:00']
        cls.probe = 'b'

        # Load validation variables from the test file
        calfile_name = f"{TEST_DATASET_PATH}validation_el{cls.probe}_state_{cls.t[0][0:4]+cls.t[0][5:7]+cls.t[0][8:10]}.tplot"
        calfile = download(remote_file=calfile_name,
                           remote_path=CONFIG['remote_data_dir'],
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not calfile:
            raise unittest.SkipTest(f"Cannot download validation file {calfile_name}")
        filename = CONFIG['local_data_dir'] + calfile_name
        tplot_restore(filename)
        cls.elf_pos_gei = pytplot.get_data(f"el{cls.probe}_pos_gei")
        cls.elf_vel_gei = pytplot.get_data(f"el{cls.probe}_vel_gei")
        cls.elf_att_gei = pytplot.get_data(f"el{cls.probe}_att_gei")
        cls.elf_att_solution = pytplot.get_data(f"el{cls.probe}_att_solution_date")
        cls.elf_att_flag = pytplot.get_data(f"el{cls.probe}_att_flag")
        cls.elf_att_spinper = pytplot.get_data(f"el{cls.probe}_att_spinper")
        cls.elf_spin_orbnorm = pytplot.get_data(f"el{cls.probe}_spin_orbnorm_angle")
        cls.elf_spin_sun = pytplot.get_data(f"el{cls.probe}_spin_sun_angle")


    def setUp(self):
        """ We need to clean tplot variables before each run"""
        pytplot.del_data('*')

    def test_state(self):
        """Validate state data."""
        pyspedas.elfin.state(trange=self.t, probe=self.probe)
        elf_pos_gei = pytplot.get_data(f"el{self.probe}_pos_gei")
        elf_vel_gei = pytplot.get_data(f"el{self.probe}_vel_gei")
        elf_att_gei = pytplot.get_data(f"el{self.probe}_att_gei")
        elf_att_solution = pytplot.get_data(f"el{self.probe}_att_solution_date")
        elf_att_flag = pytplot.get_data(f"el{self.probe}_att_flag")
        elf_att_spinper = pytplot.get_data(f"el{self.probe}_att_spinper")
        elf_spin_orbnorm = pytplot.get_data(f"el{self.probe}_spin_orbnorm_angle")
        elf_spin_sun = pytplot.get_data(f"el{self.probe}_spin_sun_angle")

        assert_array_almost_equal(elf_pos_gei.y, self.elf_pos_gei.y, decimal=4)
        assert_array_almost_equal(elf_vel_gei.y, self.elf_vel_gei.y, decimal=4)
        assert_array_almost_equal(elf_att_gei.y, self.elf_att_gei.y, decimal=2)
        assert_allclose(elf_att_solution.times, self.elf_att_solution.y, rtol=1e-3)
        assert_array_equal(elf_att_flag.y, self.elf_att_flag.y)
        assert_allclose(elf_att_spinper.y, self.elf_att_spinper.y, rtol=1e-2)
        assert_allclose(elf_spin_orbnorm.y, self.elf_spin_orbnorm.y, rtol=1e-2)
        assert_allclose(elf_spin_sun.y, self.elf_spin_sun.y, rtol=1e-2)

        logging.info("STATE DATA TEST FINISHED.")


if __name__ == '__main__':
    unittest.main()



from .tplot_average import tplot_average


def slice2d_get_support(variable, trange, matrix=False):
    """
    Retrieve user specified support data for spd_slice2d

    Input
    ------
        variable: str or float or ndarray
            Tplot variable or input data

        trange: list of float
            Time range to average over if the input is a tplot variable

    Parameters
    -----------
        matrix: ndarray
            Specifies that the input is a matrix

    Returns
    --------
        Output data (average over time range if the input is a tplot variable)
    """

    if variable is None:
        return

    if isinstance(variable, str):  # tplot variable
        return tplot_average(variable, trange)

    return variable
